@@ -3,32 +3,40 @@
 
 #include "ChatPage.hxx"
 #include "PageMeta.hxx"
-#include "screen_utils.hxx"
+#include "screen.hxx"
 #include "screen_status.hxx"
-#include "TextPage.hxx"
-#include "mpdclient.hxx"
 #include "i18n.h"
 #include "charset.hxx"
 #include "Command.hxx"
 #include "Options.hxx"
+#include "page/TextPage.hxx"
+#include "dialogs/TextInputDialog.hxx"
+#include "client/mpdclient.hxx"
 #include "util/StringAPI.hxx"
 
 #include <mpd/idle.h>
 
+#include <fmt/core.h>
+
 #include <string.h>
 #include <stdlib.h>
 
+using std::string_view_literals::operator""sv;
+
 static constexpr char chat_channel[] = "chat";
 
 class ChatPage final : public TextPage {
+	ScreenManager &screen;
+
 	unsigned last_connection_id = 0;
 	bool was_supported = false;
 
 	std::string prefix;
 
 public:
-	ChatPage(ScreenManager &_screen, WINDOW *w, Size size)
-		:TextPage(_screen, w, size) {}
+	ChatPage(ScreenManager &_screen, const Window _window)
+		:TextPage(_screen, _window, _screen.find_support),
+		 screen(_screen) {}
 
 private:
 	bool CheckChatSupport(struct mpdclient &c);
@@ -40,12 +48,15 @@ private:
 
 	void SendMessage(struct mpdclient &c, const char *msg) noexcept;
 
+	[[nodiscard]]
+	Co::InvokeTask EnterMessage(struct mpdclient &c);
+
 public:
 	/* virtual methods from class Page */
 	void Update(struct mpdclient &c, unsigned events) noexcept override;
 	bool OnCommand(struct mpdclient &c, Command cmd) override;
 
-	const char *GetTitle(char *, size_t) const noexcept override {
+	std::string_view GetTitle(std::span<char>) const noexcept override {
 		return _("Chat");
 	}
 };
@@ -62,7 +73,7 @@ ChatPage::CheckChatSupport(struct mpdclient &c)
 	last_connection_id = c.connection_id;
 
 	/* mpdclient_get_connection? */
-	if (!mpdclient_cmd_subscribe(&c, chat_channel))
+	if (!mpdclient_cmd_subscribe(c, chat_channel))
 		return (was_supported = false);
 	/* mpdclient_put_connection? */
 
@@ -70,9 +81,9 @@ ChatPage::CheckChatSupport(struct mpdclient &c)
 }
 
 static std::unique_ptr<Page>
-screen_chat_init(ScreenManager &screen, WINDOW *w, Size size)
+screen_chat_init(ScreenManager &screen, const Window window)
 {
-	return std::make_unique<ChatPage>(screen, w, size);
+	return std::make_unique<ChatPage>(screen, window);
 }
 
 void
@@ -83,8 +94,6 @@ ChatPage::ProcessMessage(const struct mpd_message &message)
 	assert(StringIsEqual(mpd_message_get_channel(&message), chat_channel));
 
 	Append(mpd_message_get_text(&message));
-
-	SetDirty();
 }
 
 void
@@ -118,7 +127,7 @@ ChatPage::GetPrefix() noexcept
 
 	if (!options.chat_prefix.empty()) {
 		/* Options are encoded in the "locale" charset */
-		prefix = LocaleToUtf8(options.chat_prefix.c_str()).c_str();
+		prefix = LocaleToUtf8{options.chat_prefix};
 		return prefix;
 	}
 
@@ -126,37 +135,45 @@ ChatPage::GetPrefix() noexcept
 	if (user == nullptr)
 		user = "nobody";
 
-	prefix = std::string("<") + user + "> ";
+	prefix = fmt::format("<{}>"sv, user);
 	return prefix;
 }
 
-void
+inline void
 ChatPage::SendMessage(struct mpdclient &c, const char *msg) noexcept
 {
-	const std::string full_msg = GetPrefix() + LocaleToUtf8(msg).c_str();
+	const std::string full_msg = fmt::format("{}{}"sv, GetPrefix(), (std::string_view)LocaleToUtf8{msg});
 
-	(void) mpdclient_cmd_send_message(&c, chat_channel, full_msg.c_str());
+	(void) mpdclient_cmd_send_message(c, chat_channel, full_msg.c_str());
+}
+
+inline Co::InvokeTask
+ChatPage::EnterMessage(struct mpdclient &c)
+{
+	const auto message = co_await TextInputDialog{
+		screen, _("Your message"),
+	};
+
+	/* the user entered an empty line */
+	if (message.empty())
+		co_return;
+
+	if (CheckChatSupport(c))
+		SendMessage(c, message.c_str());
+	else
+		Alert(_("Message could not be sent"));
 }
 
 bool
 ChatPage::OnCommand(struct mpdclient &c, Command cmd)
 {
+	CoCancel();
+
 	if (TextPage::OnCommand(c, cmd))
 		return true;
 
 	if (cmd == Command::PLAY) {
-		auto message = screen_readln(screen, _("Your message"),
-					     nullptr, nullptr, nullptr);
-
-		/* the user entered an empty line */
-		if (message.empty())
-			return true;
-
-		if (CheckChatSupport(c))
-			SendMessage(c, message.c_str());
-		else
-			screen_status_message(_("Message could not be sent"));
-
+		CoStart(EnterMessage(c));
 		return true;
 	}
 
