@@ -10,12 +10,14 @@
 #include "config.h" // IWYU pragma: keep
 #include "i18n.h"
 #include "charset.hxx"
-#include "mpdclient.hxx"
 #include "filelist.hxx"
-#include "screen_utils.hxx"
 #include "screen_client.hxx"
 #include "Command.hxx"
 #include "Options.hxx"
+#include "dialogs/YesNoDialog.hxx"
+#include "ui/Bell.hxx"
+#include "lib/fmt/ToSpan.hxx"
+#include "client/mpdclient.hxx"
 #include "util/UriUtil.hxx"
 
 #include <mpd/client.h>
@@ -29,9 +31,8 @@ class FileBrowserPage final : public FileListPage {
 	std::string current_path;
 
 public:
-	FileBrowserPage(ScreenManager &_screen, WINDOW *_w,
-			Size size) noexcept
-		:FileListPage(_screen, _w, size,
+	FileBrowserPage(ScreenManager &_screen, const Window _window) noexcept
+		:FileListPage(_screen, _screen, _window,
 			      options.list_format.c_str()) {}
 
 	bool GotoSong(struct mpdclient &c, const struct mpd_song &song) noexcept;
@@ -62,13 +63,15 @@ protected:
 
 private:
 	void HandleSave(struct mpdclient &c) noexcept;
-	void HandleDelete(struct mpdclient &c) noexcept;
+
+	[[nodiscard]]
+	Co::InvokeTask HandleDelete(struct mpdclient &c);
 
 public:
 	/* virtual methods from class Page */
 	void Update(struct mpdclient &c, unsigned events) noexcept override;
 	bool OnCommand(struct mpdclient &c, Command cmd) override;
-	const char *GetTitle(char *s, size_t size) const noexcept override;
+	std::string_view GetTitle(std::span<char> buffer) const noexcept override;
 };
 
 static void
@@ -98,7 +101,7 @@ FileBrowserPage::Reload(struct mpdclient &c) noexcept
 
 	lw.SetLength(filelist->size());
 
-	SetDirty();
+	SchedulePaint();
 }
 
 bool
@@ -172,7 +175,7 @@ FileBrowserPage::GotoSong(struct mpdclient &c, const struct mpd_song &song) noex
 		i = 0;
 
 	lw.SetCursor(i);
-	SetDirty();
+	SchedulePaint();
 	return true;
 }
 
@@ -207,20 +210,15 @@ FileBrowserPage::HandleSave(struct mpdclient &c) noexcept
 		}
 	}
 
-	if(defaultname)
-		playlist_save(screen, &c, nullptr,
-			      Utf8ToLocale(defaultname).c_str());
-	else
-		playlist_save(screen, &c, nullptr, nullptr);
+	CoStart(playlist_save(screen, c,
+			      defaultname != nullptr ? Utf8ToLocale{defaultname}.str() : std::string{}));
 }
 
-void
-FileBrowserPage::HandleDelete(struct mpdclient &c) noexcept
+inline Co::InvokeTask
+FileBrowserPage::HandleDelete(struct mpdclient &c)
 {
-	auto *connection = c.GetConnection();
-
-	if (connection == nullptr)
-		return;
+	if (!c.IsConnected())
+		co_return;
 
 	const auto range = lw.GetRange();
 	for (const unsigned i : range) {
@@ -234,8 +232,8 @@ FileBrowserPage::HandleDelete(struct mpdclient &c) noexcept
 			/* translators: the "delete" command is only possible
 			   for playlists; the user attempted to delete a song
 			   or a directory or something else */
-			screen_status_message(_("Deleting this item is not possible"));
-			screen_bell();
+			Alert(_("Deleting this item is not possible"));
+			Bell();
 			continue;
 		}
 
@@ -243,13 +241,14 @@ FileBrowserPage::HandleDelete(struct mpdclient &c) noexcept
 		char prompt[256];
 		snprintf(prompt, sizeof(prompt),
 			 _("Delete playlist %s?"),
-			 Utf8ToLocale(GetUriFilename(mpd_playlist_get_path(playlist))).c_str());
-		bool confirmed = screen_get_yesno(screen, prompt, false);
-		if (!confirmed) {
-			/* translators: a dialog was aborted by the user */
-			screen_status_message(_("Aborted"));
-			return;
-		}
+			 Utf8ToLocaleZ{GetUriFilename(mpd_playlist_get_path(playlist))}.c_str());
+
+		if (co_await YesNoDialog{screen, prompt} != YesNoResult::YES)
+			co_return;
+
+		auto *connection = c.GetConnection();
+		if (connection == nullptr)
+			throw std::runtime_error{"Not connected"};
 
 		if (!mpd_run_rm(connection, mpd_playlist_get_path(playlist))) {
 			c.HandleError();
@@ -260,18 +259,18 @@ FileBrowserPage::HandleDelete(struct mpdclient &c) noexcept
 
 		/* translators: MPD deleted the playlist, as requested by the
 		   user */
-		screen_status_message(_("Playlist deleted"));
+		Alert(_("Playlist deleted"));
 	}
 }
 
 static std::unique_ptr<Page>
-screen_file_init(ScreenManager &_screen, WINDOW *w, Size size) noexcept
+screen_file_init(ScreenManager &_screen, const Window window) noexcept
 {
-	return std::make_unique<FileBrowserPage>(_screen, w, size);
+	return std::make_unique<FileBrowserPage>(_screen, window);
 }
 
-const char *
-FileBrowserPage::GetTitle(char *str, size_t size) const noexcept
+std::string_view
+FileBrowserPage::GetTitle(std::span<char> buffer) const noexcept
 {
 	const char *path = nullptr, *prev = nullptr, *slash = current_path.c_str();
 
@@ -285,10 +284,9 @@ FileBrowserPage::GetTitle(char *str, size_t size) const noexcept
 		/* fall back to full path */
 		path = current_path.c_str();
 
-	snprintf(str, size, "%s: %s",
-		 /* translators: caption of the browser screen */
-		 _("Browse"), Utf8ToLocale(path).c_str());
-	return str;
+	return FmtTruncate(buffer, "{}: {}",
+			   /* translators: caption of the browser screen */
+			   _("Browse"), (std::string_view)Utf8ToLocale{path});
 }
 
 void
@@ -305,13 +303,15 @@ FileBrowserPage::Update(struct mpdclient &c, unsigned events) noexcept
 #endif
 		      )) {
 		screen_browser_sync_highlights(*filelist, c.playlist);
-		SetDirty();
+		SchedulePaint();
 	}
 }
 
 bool
 FileBrowserPage::OnCommand(struct mpdclient &c, Command cmd)
 {
+	CoCancel();
+
 	switch(cmd) {
 	case Command::GO_ROOT_DIRECTORY:
 		ChangeDirectory(c, {});
@@ -320,12 +320,6 @@ FileBrowserPage::OnCommand(struct mpdclient &c, Command cmd)
 		ChangeToParent(c);
 		return true;
 
-	case Command::LOCATE:
-		/* don't let browser_cmd() evaluate the locate command
-		   - it's a no-op, and by the way, leads to a
-		   segmentation fault in the current implementation */
-		return false;
-
 	case Command::SCREEN_UPDATE:
 		Reload(c);
 		screen_browser_sync_highlights(*filelist, c.playlist);
@@ -338,20 +332,20 @@ FileBrowserPage::OnCommand(struct mpdclient &c, Command cmd)
 	if (FileListPage::OnCommand(c, cmd))
 		return true;
 
-	if (!c.IsConnected())
+	if (!c.IsReady())
 		return false;
 
 	switch(cmd) {
 	case Command::DELETE:
-		HandleDelete(c);
-		break;
+		CoStart(HandleDelete(c));
+		return true;
 
 	case Command::SAVE_PLAYLIST:
 		HandleSave(c);
 		break;
 
 	case Command::DB_UPDATE:
-		screen_database_update(&c, current_path.c_str());
+		screen_database_update(c, current_path.c_str());
 		return true;
 
 	default:
