@@ -4,39 +4,36 @@
 #include "config.h"
 #include "FileListPage.hxx"
 #include "FileBrowserPage.hxx"
-#include "SongPage.hxx"
 #include "EditPlaylistPage.hxx"
-#include "LyricsPage.hxx"
 #include "Command.hxx"
-#include "screen_status.hxx"
-#include "screen_find.hxx"
 #include "screen.hxx"
 #include "i18n.h"
 #include "Options.hxx"
 #include "charset.hxx"
 #include "strfsong.hxx"
-#include "mpdclient.hxx"
 #include "filelist.hxx"
 #include "Styles.hxx"
-#include "paint.hxx"
 #include "SongRowPaint.hxx"
 #include "time_format.hxx"
+#include "ui/paint.hxx"
+#include "client/mpdclient.hxx"
 #include "util/UriUtil.hxx"
 
 #include <mpd/client.h>
 
 #include <string.h>
 
+using std::string_view_literals::operator""sv;
+
 #define BUFSIZE 1024
 
 #ifndef NCMPC_MINI
 static constexpr unsigned HIGHLIGHT = 0x01;
 #endif
 
-FileListPage::FileListPage(ScreenManager &_screen, WINDOW *_w,
-			   Size size,
+FileListPage::FileListPage(PageContainer &_parent, ScreenManager &_screen, Window _window,
 			   const char *_song_format) noexcept
-	:ListPage(_w, size),
+	:ListPage(_parent, _window),
 	 screen(_screen),
 	 song_format(_song_format)
 {
@@ -67,8 +64,8 @@ screen_browser_sync_highlights(FileList &fl, const MpdQueue &playlist) noexcept
 
 #endif
 
-const char *
-FileListPage::GetListItemText(char *buffer, size_t size,
+std::string_view
+FileListPage::GetListItemText(std::span<char> buffer,
 			      unsigned idx) const noexcept
 {
 	assert(filelist != nullptr);
@@ -78,80 +75,82 @@ FileListPage::GetListItemText(char *buffer, size_t size,
 	const auto *entity = entry.entity;
 
 	if( entity == nullptr )
-		return "..";
+		return ".."sv;
 
 	if (mpd_entity_get_type(entity) == MPD_ENTITY_TYPE_DIRECTORY) {
 		const auto *dir = mpd_entity_get_directory(entity);
 		const char *name = GetUriFilename(mpd_directory_get_path(dir));
-		return utf8_to_locale(name, buffer, size);
+		return utf8_to_locale(name, buffer);
 	} else if (mpd_entity_get_type(entity) == MPD_ENTITY_TYPE_SONG) {
 		const auto *song = mpd_entity_get_song(entity);
 
-		strfsong(buffer, size, options.list_format.c_str(), song);
-		return buffer;
+		return strfsong(buffer, options.list_format.c_str(), *song);
 	} else if (mpd_entity_get_type(entity) == MPD_ENTITY_TYPE_PLAYLIST) {
 		const auto *playlist = mpd_entity_get_playlist(entity);
 		const char *name = GetUriFilename(mpd_playlist_get_path(playlist));
-		return utf8_to_locale(name, buffer, size);
+		return utf8_to_locale(name, buffer);
 	}
 
-	return "Error: Unknown entry!";
+	return "Error: Unknown entry!"sv;
 }
 
-static bool
-load_playlist(struct mpdclient *c, const struct mpd_playlist *playlist)
+bool
+FileListPage::LoadPlaylist(struct mpdclient &c,
+			   const struct mpd_playlist &playlist) noexcept
 {
-	auto *connection = c->GetConnection();
+	auto *connection = c.GetConnection();
 	if (connection == nullptr)
 		return false;
 
-	if (mpd_run_load(connection, mpd_playlist_get_path(playlist))) {
-		const char *name = GetUriFilename(mpd_playlist_get_path(playlist));
-		screen_status_printf(_("Loading playlist '%s'"),
-				     Utf8ToLocale(name).c_str());
+	if (mpd_run_load(connection, mpd_playlist_get_path(&playlist))) {
+		const char *name = GetUriFilename(mpd_playlist_get_path(&playlist));
+		FmtAlert(_("Loading playlist '{}'"),
+			 Utf8ToLocaleZ{name}.c_str());
 
-		c->events |= MPD_IDLE_QUEUE;
+		c.events |= MPD_IDLE_QUEUE;
 	} else
-		c->HandleError();
+		c.HandleError();
 
 	return true;
 }
 
-static bool
-enqueue_and_play(struct mpdclient *c, FileListEntry *entry)
+inline bool
+FileListPage::EnqueueAndPlay(struct mpdclient &c, FileListEntry &entry) noexcept
 {
-	auto *connection = c->GetConnection();
+	auto *connection = c.GetConnection();
 	if (connection == nullptr)
 		return false;
 
-	const auto *song = mpd_entity_get_song(entry->entity);
+	const auto *song = mpd_entity_get_song(entry.entity);
+	assert(song != nullptr);
+
 	int id;
 
 #ifndef NCMPC_MINI
-	if (!(entry->flags & HIGHLIGHT))
+	if (!(entry.flags & HIGHLIGHT))
 		id = -1;
 	else
 #endif
-		id = c->playlist.FindIdByUri(mpd_song_get_uri(song));
+		id = c.playlist.FindIdByUri(mpd_song_get_uri(song));
 
 	if (id < 0) {
-		char buf[BUFSIZE];
-
 		id = mpd_run_add_id(connection, mpd_song_get_uri(song));
 		if (id < 0) {
-			c->HandleError();
+			c.HandleError();
 			return false;
 		}
 
 #ifndef NCMPC_MINI
-		entry->flags |= HIGHLIGHT;
+		entry.flags |= HIGHLIGHT;
 #endif
-		strfsong(buf, BUFSIZE, options.list_format.c_str(), song);
-		screen_status_printf(_("Adding \'%s\' to queue"), buf);
+
+		char buf[BUFSIZE];
+		FmtAlert(_("Adding '{}' to queue"),
+			 strfsong(buf, options.list_format.c_str(), *song));
 	}
 
 	if (!mpd_run_play_id(connection, id)) {
-		c->HandleError();
+		c.HandleError();
 		return false;
 	}
 
@@ -214,29 +213,29 @@ FileListPage::HandleEnter(struct mpdclient &c)
 		return false;
 
 	if (mpd_entity_get_type(entity) == MPD_ENTITY_TYPE_PLAYLIST)
-		return load_playlist(&c, mpd_entity_get_playlist(entity));
+		return LoadPlaylist(c, *mpd_entity_get_playlist(entity));
 	else if (mpd_entity_get_type(entity) == MPD_ENTITY_TYPE_SONG)
-		return enqueue_and_play(&c, entry);
+		return EnqueueAndPlay(c, *entry);
 	return false;
 }
 
-static bool
-browser_select_entry(struct mpdclient &c, FileListEntry &entry,
-		     [[maybe_unused]] bool toggle) noexcept
+bool
+FileListPage::HandleSelectEntry(struct mpdclient &c, FileListEntry &entry,
+				[[maybe_unused]] bool toggle) noexcept
 {
 	assert(entry.entity != nullptr);
 
 	if (mpd_entity_get_type(entry.entity) == MPD_ENTITY_TYPE_PLAYLIST)
-		return load_playlist(&c, mpd_entity_get_playlist(entry.entity));
+		return LoadPlaylist(c, *mpd_entity_get_playlist(entry.entity));
 
 	if (mpd_entity_get_type(entry.entity) == MPD_ENTITY_TYPE_DIRECTORY) {
 		const auto *dir = mpd_entity_get_directory(entry.entity);
 
-		if (!mpdclient_cmd_add_path(&c, mpd_directory_get_path(dir)))
+		if (!mpdclient_cmd_add_path(c, mpd_directory_get_path(dir)))
 			return false;
 
-		screen_status_printf(_("Adding \'%s\' to queue"),
-				     Utf8ToLocale(mpd_directory_get_path(dir)).c_str());
+		FmtAlert(_("Adding '{}' to queue"),
+			 (std::string_view)Utf8ToLocale{mpd_directory_get_path(dir)});
 		return true;
 	}
 
@@ -255,10 +254,8 @@ browser_select_entry(struct mpdclient &c, FileListEntry &entry,
 
 		if (c.RunAdd(*song)) {
 			char buf[BUFSIZE];
-
-			strfsong(buf, BUFSIZE,
-				 options.list_format.c_str(), song);
-			screen_status_printf(_("Adding \'%s\' to queue"), buf);
+			FmtAlert(_("Adding '{}' to queue"),
+				 strfsong(buf, options.list_format.c_str(), *song));
 		}
 #ifndef NCMPC_MINI
 	} else {
@@ -285,10 +282,10 @@ FileListPage::HandleSelect(struct mpdclient &c) noexcept
 	for (const unsigned i : range) {
 		auto *entry = GetIndex(i);
 		if (entry != nullptr && entry->entity != nullptr)
-			success = browser_select_entry(c, *entry, true);
+			success = HandleSelectEntry(c, *entry, true);
 	}
 
-	SetDirty();
+	SchedulePaint();
 
 	return range.end_index == range.start_index + 1 && success;
 }
@@ -302,7 +299,7 @@ FileListPage::HandleAdd(struct mpdclient &c) noexcept
 	for (const unsigned i : range) {
 		auto *entry = GetIndex(i);
 		if (entry != nullptr && entry->entity != nullptr)
-			success = browser_select_entry(c, *entry, false) ||
+			success = HandleSelectEntry(c, *entry, false) ||
 				success;
 	}
 
@@ -338,10 +335,10 @@ FileListPage::HandleSelectAll(struct mpdclient &c) noexcept
 		auto &entry = (*filelist)[i];
 
 		if (entry.entity != nullptr)
-			browser_select_entry(c, entry, false);
+			HandleSelectEntry(c, entry, false);
 	}
 
-	SetDirty();
+	SchedulePaint();
 }
 
 #ifdef HAVE_GETMOUSE
@@ -367,7 +364,7 @@ FileListPage::OnMouse(struct mpdclient &c, Point p,
 			HandleSelect(c);
 	}
 
-	SetDirty();
+	SchedulePaint();
 
 	return true;
 }
@@ -388,42 +385,17 @@ FileListPage::OnCommand(struct mpdclient &c, Command cmd)
 	case Command::LIST_RFIND:
 	case Command::LIST_FIND_NEXT:
 	case Command::LIST_RFIND_NEXT:
-		screen_find(screen, lw, cmd, *this);
-		SetDirty();
+		CoStart(screen.find_support.Find(lw, *this, cmd));
 		return true;
 	case Command::LIST_JUMP:
-		screen_jump(screen, lw, *this, *this);
-		SetDirty();
-		return true;
-
-#ifdef ENABLE_SONG_SCREEN
-	case Command::SCREEN_SONG:
-		if (const auto *song = GetSelectedSong()) {
-			screen_song_switch(screen, c, *song);
-			return true;
-		} else
-			return false;
-
-#endif
-
-#ifdef ENABLE_LYRICS_SCREEN
-	case Command::SCREEN_LYRICS:
-		if (const auto *song = GetSelectedSong()) {
-			screen_lyrics_switch(screen, c, *song, false);
-			return true;
-		} else
-			return false;
-
-#endif
-	case Command::SCREEN_SWAP:
-		screen.Swap(c, GetSelectedSong());
+		CoStart(screen.find_support.Jump(lw, *this, *this));
 		return true;
 
 	default:
 		break;
 	}
 
-	if (!c.IsConnected())
+	if (!c.IsReady())
 		return false;
 
 	switch (cmd) {
@@ -434,13 +406,13 @@ FileListPage::OnCommand(struct mpdclient &c, Command cmd)
 	case Command::SELECT:
 		if (HandleSelect(c))
 			lw.HandleCommand(Command::LIST_NEXT);
-		SetDirty();
+		SchedulePaint();
 		return true;
 
 	case Command::ADD:
 		if (HandleAdd(c))
 			lw.HandleCommand(Command::LIST_NEXT);
-		SetDirty();
+		SchedulePaint();
 		return true;
 
 #ifdef ENABLE_PLAYLIST_EDITOR
@@ -454,14 +426,6 @@ FileListPage::OnCommand(struct mpdclient &c, Command cmd)
 		HandleSelectAll(c);
 		return true;
 
-	case Command::LOCATE:
-		if (const auto *song = GetSelectedSong()) {
-			screen_file_goto_song(screen, c, *song);
-			return true;
-		} else
-			return false;
-
-
 	default:
 		break;
 	}
@@ -470,28 +434,28 @@ FileListPage::OnCommand(struct mpdclient &c, Command cmd)
 }
 
 void
-screen_browser_paint_directory(WINDOW *w, unsigned width,
-			       bool selected, const char *name) noexcept
+screen_browser_paint_directory(const Window window, unsigned width,
+			       bool selected, std::string_view name) noexcept
 {
-	row_color(w, Style::DIRECTORY, selected);
+	row_color(window, Style::DIRECTORY, selected);
 
-	waddch(w, '[');
-	waddstr(w, name);
-	waddch(w, ']');
+	window.Char('[');
+	window.String(name);
+	window.Char(']');
 
 	/* erase the unused space after the text */
-	row_clear_to_eol(w, width, selected);
+	row_clear_to_eol(window, width, selected);
 }
 
 static void
-screen_browser_paint_playlist(WINDOW *w, unsigned width,
-			      bool selected, const char *name) noexcept
+screen_browser_paint_playlist(const Window window, unsigned width,
+			      bool selected, std::string_view name) noexcept
 {
-	row_paint_text(w, width, Style::PLAYLIST, selected, name);
+	row_paint_text(window, width, Style::PLAYLIST, selected, name);
 }
 
 void
-FileListPage::PaintListItem(WINDOW *w, unsigned i,
+FileListPage::PaintListItem(const Window window, unsigned i,
 			    unsigned y, unsigned width,
 			    bool selected) const noexcept
 {
@@ -501,7 +465,7 @@ FileListPage::PaintListItem(WINDOW *w, unsigned i,
 	const auto &entry = (*filelist)[i];
 	const struct mpd_entity *entity = entry.entity;
 	if (entity == nullptr) {
-		screen_browser_paint_directory(w, width, selected, "..");
+		screen_browser_paint_directory(window, width, selected, ".."sv);
 		return;
 	}
 
@@ -515,27 +479,25 @@ FileListPage::PaintListItem(WINDOW *w, unsigned i,
 	case MPD_ENTITY_TYPE_DIRECTORY: {
 		const auto *directory = mpd_entity_get_directory(entity);
 		const char *name = GetUriFilename(mpd_directory_get_path(directory));
-		screen_browser_paint_directory(w, width, selected,
-					       Utf8ToLocale(name).c_str());
+		screen_browser_paint_directory(window, width, selected, Utf8ToLocale{name});
 		break;
 	}
 
 	case MPD_ENTITY_TYPE_SONG:
-		paint_song_row(w, y, width, selected, highlight,
-			       mpd_entity_get_song(entity), nullptr,
+		paint_song_row(window, y, width, selected, highlight,
+			       *mpd_entity_get_song(entity), nullptr,
 			       song_format);
 		break;
 
 	case MPD_ENTITY_TYPE_PLAYLIST: {
 		const auto *playlist = mpd_entity_get_playlist(entity);
 		const char *name = GetUriFilename(mpd_playlist_get_path(playlist));
-		screen_browser_paint_playlist(w, width, selected,
-					      Utf8ToLocale(name).c_str());
+		screen_browser_paint_playlist(window, width, selected, Utf8ToLocale{name});
 		break;
 	}
 
 	default:
-		row_paint_text(w, width,
+		row_paint_text(window, width,
 			       highlight ? Style::LIST_BOLD : Style::LIST,
 			       selected, "<unknown>");
 	}
@@ -548,15 +510,13 @@ FileListPage::Paint() const noexcept
 }
 
 bool
-FileListPage::PaintStatusBarOverride(const Window &window) const noexcept
+FileListPage::PaintStatusBarOverride(const Window window) const noexcept
 {
 	if (!lw.HasRangeSelection())
 		return false;
 
-	WINDOW *const w = window.w;
-
-	wmove(w, 0, 0);
-	wclrtoeol(w);
+	window.MoveCursor({0, 0});
+	window.ClearToEol();
 
 	unsigned duration = 0;
 
@@ -571,15 +531,14 @@ FileListPage::PaintStatusBarOverride(const Window &window) const noexcept
 			duration += mpd_song_get_duration(mpd_entity_get_song(entity));
 	}
 
-	char duration_string[32];
-	format_duration_short(duration_string, sizeof(duration_string),
-			      duration);
-	const unsigned duration_width = strlen(duration_string);
+	char duration_buffer[32];
+	const auto duration_string = format_duration_short(duration_buffer, duration);
+	const unsigned duration_width = duration_string.size();
 
-	SelectStyle(w, Style::STATUS_TIME);
-	mvwaddstr(w, 0, window.size.width - duration_width, duration_string);
+	SelectStyle(window, Style::STATUS_TIME);
+	window.String({0, (int)window.GetWidth() - (int)duration_width}, duration_string);
 
-	wnoutrefresh(w);
+	window.RefreshNoOut();
 
 	return true;
 }
