@@ -1,30 +1,13 @@
-/* ncmpc (Ncurses MPD Client)
- * (c) 2004-2020 The Music Player Daemon Project
- * Project homepage: http://musicpd.org
- *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License along
- * with this program; if not, write to the Free Software Foundation, Inc.,
- * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
- */
+// SPDX-License-Identifier: GPL-2.0-or-later
+// Copyright The Music Player Daemon Project
 
 #include "Instance.hxx"
 #include "Options.hxx"
 #include "event/SignalMonitor.hxx"
+#include "util/Exception.hxx" // for GetFullMessage()
 
 #include <signal.h>
 
-#ifdef HAVE_TAG_WHITELIST
-
 #include "strfsong.hxx"
 #include "TagMask.hxx"
 
@@ -34,23 +17,31 @@ static constexpr TagMask global_tag_whitelist{
 	MPD_TAG_TITLE,
 	MPD_TAG_ALBUM,
 	MPD_TAG_COMPOSER,
+	MPD_TAG_PERFORMER,
+#if LIBMPDCLIENT_CHECK_VERSION(2,17,0)
+	MPD_TAG_CONDUCTOR,
+	MPD_TAG_WORK,
+#endif
 	MPD_TAG_NAME,
 	MPD_TAG_DISC,
 	MPD_TAG_TRACK,
 	MPD_TAG_DATE,
 	MPD_TAG_GENRE,
+#if LIBMPDCLIENT_CHECK_VERSION(2,17,0)
+	MPD_TAG_LABEL,
+#endif
 	MPD_TAG_COMMENT,
 };
 
-#endif
-
 Instance::Instance()
 	:client(event_loop,
 		options.host.empty() ? nullptr : options.host.c_str(),
 		options.port,
 		options.timeout_ms,
-		options.password.empty() ? nullptr : options.password.c_str()),
+		options.password.empty() ? nullptr : options.password.c_str(),
+		*this),
 	 seek(event_loop, client),
+	 update_screen_event(event_loop, BIND_THIS_METHOD(OnUpdateScreen)),
 	 reconnect_timer(event_loop, BIND_THIS_METHOD(OnReconnectTimer)),
 	 update_timer(event_loop, BIND_THIS_METHOD(OnUpdateTimer)),
 #ifndef NCMPC_MINI
@@ -58,9 +49,9 @@ Instance::Instance()
 #endif
 	 screen_manager(event_loop),
 #ifdef ENABLE_LIRC
-	 lirc_input(event_loop),
+	 lirc_input(event_loop, *this),
 #endif
-	 user_input(event_loop, *screen_manager.main_window.w)
+	 user_input(event_loop, *screen_manager.main_window.w, *this)
 {
 	screen_manager.Init(&client);
 
@@ -73,7 +64,6 @@ Instance::Instance()
 	SignalMonitorRegister(SIGCONT, BIND_THIS_METHOD(OnSigwinch));
 #endif
 
-#ifdef HAVE_TAG_WHITELIST
 	TagMask tag_mask = global_tag_whitelist;
 	tag_mask |= SongFormatToTagMask(options.list_format.c_str());
 	tag_mask |= SongFormatToTagMask(options.search_format.c_str());
@@ -83,7 +73,6 @@ Instance::Instance()
 #endif
 
 	client.WhitelistTags(tag_mask);
-#endif
 }
 
 Instance::~Instance()
@@ -99,3 +88,20 @@ Instance::Run()
 
 	event_loop.Run();
 }
+
+bool
+Instance::OnRawKey(int key) noexcept
+try {
+	update_screen_event.Schedule();
+
+	return screen_manager.OnModalDialogKey(key);
+} catch (...) {
+	screen_manager.status_bar.SetMessage(GetFullMessage(std::current_exception()));
+	return true;
+}
+
+bool
+Instance::CancelModalDialog() noexcept
+{
+	return screen_manager.CancelModalDialog();
+}
