@@ -6,28 +6,32 @@
 #include "LyricsLoader.hxx"
 #include "PageMeta.hxx"
 #include "screen_status.hxx"
-#include "FileBrowserPage.hxx"
-#include "SongPage.hxx"
 #include "i18n.h"
 #include "Command.hxx"
 #include "Options.hxx"
-#include "mpdclient.hxx"
 #include "screen.hxx"
 #include "plugin.hxx"
-#include "TextPage.hxx"
-#include "screen_utils.hxx"
-#include "ncu.hxx"
+#include "page/TextPage.hxx"
+#include "lib/fmt/ToSpan.hxx"
+#include "client/mpdclient.hxx"
 #include "util/StringAPI.hxx"
 
 #include <string>
 
 #include <assert.h>
 #include <errno.h>
+#include <spawn.h>
 #include <sys/wait.h>
 #include <stdlib.h>
 #include <string.h>
 #include <stdio.h>
 
+using std::string_view_literals::operator""sv;
+
+#ifdef __APPLE__
+extern char **environ;
+#endif
+
 static struct mpd_song *next_song;
 static bool follow = false;
 
@@ -54,12 +58,14 @@ class LyricsPage final : public TextPage, PluginResponseHandler {
 	CoarseTimerEvent plugin_timeout;
 
 public:
-	LyricsPage(ScreenManager &_screen, WINDOW *w, Size size)
-		:TextPage(_screen, w, size),
-		 plugin_timeout(_screen.GetEventLoop(),
+	LyricsPage(EventLoop &event_loop,
+		   PageContainer &_parent, const Window _window,
+		   FindSupport &_find_support) noexcept
+		:TextPage(_parent, _window, _find_support),
+		 plugin_timeout(event_loop,
 				BIND_THIS_METHOD(OnTimeout)) {}
 
-	~LyricsPage() override {
+	~LyricsPage() noexcept override {
 		Cancel();
 	}
 
@@ -75,19 +81,9 @@ private:
 		plugin_cycle = nullptr;
 	}
 
-	void Cancel();
+	void Cancel() noexcept;
 
-	/**
-	 * Repaint and update the screen, if it is currently active.
-	 */
-	void RepaintIfActive() {
-		if (screen.IsVisible(*this))
-			Repaint();
-
-		/* XXX repaint the screen title */
-	}
-
-	void Set(const char *s);
+	void Set(const char *s) noexcept;
 
 	void StartPluginCycle() noexcept;
 
@@ -99,12 +95,12 @@ private:
 			MaybeLoad(*new_song);
 	}
 
-	void Reload();
+	void Reload() noexcept;
 
-	bool Save();
+	bool Save() noexcept;
 
 	/** save current lyrics to a file and run editor on it */
-	void Edit();
+	void Edit() noexcept;
 
 	void OnTimeout() noexcept;
 
@@ -113,7 +109,11 @@ public:
 	void OnOpen(struct mpdclient &c) noexcept override;
 	void Update(struct mpdclient &c, unsigned events) noexcept override;
 	bool OnCommand(struct mpdclient &c, Command cmd) override;
-	const char *GetTitle(char *, size_t) const noexcept override;
+	std::string_view GetTitle(std::span<char> buffer) const noexcept override;
+
+	const struct mpd_song *GetSelectedSong() const noexcept override {
+		return song;
+	}
 
 private:
 	/* virtual methods from class PluginResponseHandler */
@@ -123,7 +123,7 @@ private:
 };
 
 void
-LyricsPage::Cancel()
+LyricsPage::Cancel() noexcept
 {
 	if (plugin_cycle != nullptr)
 		StopPluginCycle();
@@ -141,7 +141,7 @@ LyricsPage::Cancel()
 }
 
 bool
-LyricsPage::Save()
+LyricsPage::Save() noexcept
 {
 	FILE *lyr_file = cache.Save(artist, title);
 	if (lyr_file == nullptr)
@@ -155,7 +155,7 @@ LyricsPage::Save()
 }
 
 void
-LyricsPage::Set(const char *s)
+LyricsPage::Set(const char *s) noexcept
 {
 	if (reloading) {
 		unsigned saved_start = lw.GetOrigin();
@@ -170,10 +170,6 @@ LyricsPage::Set(const char *s)
 	}
 
 	reloading = false;
-
-	/* paint new data */
-
-	RepaintIfActive();
 }
 
 void
@@ -193,7 +189,7 @@ LyricsPage::OnPluginSuccess(const char *_plugin_name,
 	StopPluginCycle();
 
 	/* schedule a full repaint so the page title gets updated */
-	screen.SchedulePaint();
+	SchedulePaint();
 }
 
 void
@@ -204,13 +200,13 @@ LyricsPage::OnPluginError(std::string error) noexcept
 	Set(error.c_str());
 
 	/* translators: no lyrics were found for the song */
-	screen_status_message(_("No lyrics"));
+	Alert(_("No lyrics"));
 
 	plugin_timeout.Cancel();
 	StopPluginCycle();
 
 	/* schedule a full repaint so the page title gets updated */
-	screen.SchedulePaint();
+	SchedulePaint();
 }
 
 void
@@ -218,11 +214,11 @@ LyricsPage::OnTimeout() noexcept
 {
 	StopPluginCycle();
 
-	screen_status_printf(_("Lyrics timeout occurred after %d seconds"),
-			     (int)std::chrono::duration_cast<std::chrono::seconds>(options.lyrics_timeout).count());
+	FmtAlert(_("Lyrics timeout occurred after {} seconds"),
+		 (int)std::chrono::duration_cast<std::chrono::seconds>(options.lyrics_timeout).count());
 
 	/* schedule a full repaint so the page title gets updated */
-	screen.SchedulePaint();
+	SchedulePaint();
 }
 
 void
@@ -272,19 +268,21 @@ LyricsPage::MaybeLoad(const struct mpd_song &new_song) noexcept
 }
 
 void
-LyricsPage::Reload()
+LyricsPage::Reload() noexcept
 {
 	if (plugin_cycle == nullptr && artist != nullptr && title != nullptr) {
 		reloading = true;
 		StartPluginCycle();
-		Repaint();
+		SchedulePaint();
 	}
 }
 
 static std::unique_ptr<Page>
-lyrics_screen_init(ScreenManager &_screen, WINDOW *w, Size size)
+lyrics_screen_init(ScreenManager &_screen, const Window window) noexcept
 {
-	return std::make_unique<LyricsPage>(_screen, w, size);
+	return std::make_unique<LyricsPage>(_screen.GetEventLoop(),
+					    _screen, window,
+					    _screen.find_support);
 }
 
 void
@@ -308,82 +306,81 @@ LyricsPage::Update(struct mpdclient &c, unsigned) noexcept
 		MaybeLoad(c.GetPlayingSong());
 }
 
-const char *
-LyricsPage::GetTitle(char *str, size_t size) const noexcept
+std::string_view
+LyricsPage::GetTitle(std::span<char> buffer) const noexcept
 {
 	if (plugin_cycle != nullptr) {
-		snprintf(str, size, "%s (%s)",
-			 _("Lyrics"),
-			 /* translators: this message is displayed
-			    while data is retrieved */
-			 _("loading..."));
-		return str;
+		return FmtTruncate(buffer, "{} ({})",
+				   _("Lyrics"),
+				   /* translators: this message is
+				      displayed while data is
+				      retrieved */
+				   _("loading..."));
 	} else if (artist != nullptr && title != nullptr && !IsEmpty()) {
-		int n;
-		n = snprintf(str, size, "%s: %s - %s",
+		std::size_t n;
+		n = snprintf(buffer.data(), buffer.size(), "%s: %s - %s",
 			     _("Lyrics"),
 			     artist, title);
 
 		if (options.lyrics_show_plugin && !plugin_name.empty() &&
-		    (unsigned int) n < size - 1)
-			snprintf(str + n, size - n, " (%s)",
-				 plugin_name.c_str());
+		    n < buffer.size() - 1)
+			n += snprintf(buffer.data() + n, buffer.size() - n, " (%s)", plugin_name.c_str());
 
-		return str;
+		return {buffer.data(), n};
 	} else
 		return _("Lyrics");
 }
 
 void
-LyricsPage::Edit()
+LyricsPage::Edit() noexcept
 {
 	const auto path = cache.MakePath(artist, title);
 	if (path.empty()) {
-		screen_status_message(_("Lyrics cache is unavailable"));
+		Alert(_("Lyrics cache is unavailable"));
 		return;
 	}
 
 	if (options.text_editor.empty()) {
-		screen_status_message(_("Editor not configured"));
+		Alert(_("Editor not configured"));
 		return;
 	}
 
 	const char *editor = options.text_editor.c_str();
-	if (options.text_editor_ask) {
-		const char *prompt =
-			_("Do you really want to start an editor and edit these lyrics?");
-		bool really = screen_get_yesno(screen, prompt, false);
-		if (!really) {
-			screen_status_message(_("Aborted"));
-			return;
-		}
-	}
 
 	if (!Save())
 		return;
 
-	ncu_deinit();
+	def_prog_mode();
+	endwin();
 
-	/* TODO: fork/exec/wait won't work on Windows, but building a command
-	   string for system() is too tricky */
-	int status;
-	pid_t pid = fork();
-	if (pid == -1) {
-		screen_status_printf(("%s (%s)"), _("Can't start editor"), strerror(errno));
-		ncu_init();
+	posix_spawnattr_t attr;
+	posix_spawnattr_init(&attr);
+
+#ifdef USE_SIGNALFD
+	/* unblock all signals which may be blocked for signalfd */
+	posix_spawnattr_setflags(&attr, POSIX_SPAWN_SETSIGMASK);
+#endif
+
+	char *const argv[] = {
+		const_cast<char *>(editor),
+		const_cast<char *>(path.c_str()),
+		nullptr
+	};
+
+	pid_t pid;
+	if (posix_spawn(&pid, editor, nullptr, &attr,
+			argv, environ) != 0) {
+		reset_prog_mode();
+		FmtAlert("{} ({})"sv, _("Can't start editor"), strerror(errno));
 		return;
-	} else if (pid == 0) {
-		execlp(editor, editor, path.c_str(), nullptr);
-		/* exec failed, do what system does */
-		_exit(127);
-	} else {
-		int ret;
-		do {
-			ret = waitpid(pid, &status, 0);
-		} while (ret == -1 && errno == EINTR);
 	}
 
-	ncu_init();
+	int ret, status;
+	do {
+		ret = waitpid(pid, &status, 0);
+	} while (ret == -1 && errno == EINTR);
+
+	reset_prog_mode();
 
 	/* TODO: hardly portable */
 	if (WIFEXITED(status)) {
@@ -391,15 +388,15 @@ LyricsPage::Edit()
 			/* update to get the changes */
 			Reload();
 		else if (WEXITSTATUS(status) == 127)
-			screen_status_message(_("Can't start editor"));
+			Alert(_("Can't start editor"));
 		else
-			screen_status_printf("%s (%d)",
-					     _("Editor exited unexpectedly"),
-					     WEXITSTATUS(status));
+			FmtAlert("{} ({})",
+				 _("Editor exited unexpectedly"),
+				 WEXITSTATUS(status));
 	} else if (WIFSIGNALED(status)) {
-		screen_status_printf("%s (signal %d)",
-				     _("Editor exited unexpectedly"),
-				     WTERMSIG(status));
+		FmtAlert("{} (signal {})"sv,
+			 _("Editor exited unexpectedly"),
+			 WTERMSIG(status));
 	}
 }
 
@@ -420,14 +417,14 @@ LyricsPage::OnCommand(struct mpdclient &c, Command cmd)
 		if (plugin_cycle == nullptr && artist != nullptr &&
 		    title != nullptr && Save())
 			/* lyrics for the song were saved on hard disk */
-			screen_status_message (_("Lyrics saved"));
+			Alert (_("Lyrics saved"));
 		return true;
 	case Command::DELETE:
 		if (plugin_cycle == nullptr && artist != nullptr &&
 		    title != nullptr) {
-			screen_status_message(cache.Delete(artist, title)
-					      ? _("Lyrics deleted")
-					      : _("No saved lyrics"));
+			Alert(cache.Delete(artist, title)
+			      ? _("Lyrics deleted")
+			      : _("No saved lyrics"));
 		}
 		return true;
 	case Command::LYRICS_UPDATE:
@@ -441,27 +438,6 @@ LyricsPage::OnCommand(struct mpdclient &c, Command cmd)
 		Reload();
 		return true;
 
-#ifdef ENABLE_SONG_SCREEN
-	case Command::SCREEN_SONG:
-		if (song != nullptr) {
-			screen_song_switch(screen, c, *song);
-			return true;
-		}
-
-		break;
-#endif
-	case Command::SCREEN_SWAP:
-		screen.Swap(c, song);
-		return true;
-
-	case Command::LOCATE:
-		if (song != nullptr) {
-			screen_file_goto_song(screen, c, *song);
-			return true;
-		}
-
-		return false;
-
 	default:
 		break;
 	}
