@@ -4,8 +4,6 @@
 #include "config.h"
 #include "Instance.hxx"
 #include "ncmpc.hxx"
-#include "mpdclient.hxx"
-#include "callbacks.hxx"
 #include "charset.hxx"
 #include "Options.hxx"
 #include "Command.hxx"
@@ -38,7 +36,7 @@
 #include <locale.h>
 #endif
 
-static Instance *global_instance;
+using std::string_view_literals::operator""sv;
 
 ScreenManager *screen;
 
@@ -51,20 +49,20 @@ update_xterm_title(struct mpdclient &client) noexcept
 	static constexpr std::size_t BUFSIZE = 1024;
 
 	char tmp[BUFSIZE];
-	const char *new_title = nullptr;
-	if (!options.xterm_title_format.empty() && song != nullptr)
-		new_title = strfsong(tmp, BUFSIZE,
-				     options.xterm_title_format.c_str(), song) > 0
-			? tmp
-			: nullptr;
+	std::string_view new_title = PACKAGE " version " VERSION ""sv;
+	if (!options.xterm_title_format.empty() && song != nullptr) {
+		const auto s = strfsong(tmp, options.xterm_title_format.c_str(), *song);
+		if (!s.empty())
+			new_title = s;
+	}
 
-	if (new_title == nullptr)
-		new_title = PACKAGE " version " VERSION;
+	static char old_title[BUFSIZE];
+	static std::size_t old_title_length = 0;
 
-	static char title[BUFSIZE];
-	if (strncmp(title, new_title, BUFSIZE)) {
-		strcpy(title, new_title);
-		set_xterm_title(title);
+	if (new_title != std::string_view{old_title, old_title_length}) {
+		std::copy(new_title.begin(), new_title.end(), old_title);
+		old_title_length = new_title.size();
+		set_xterm_title(new_title);
 	}
 }
 #endif
@@ -87,7 +85,7 @@ auto_update_timer(Instance &instance) noexcept
 void
 Instance::UpdateClient() noexcept
 {
-	if (client.IsConnected() &&
+	if (client.IsReady() &&
 	    (client.events != 0 || client.playing))
 		client.Update();
 
@@ -113,11 +111,10 @@ Instance::OnReconnectTimer() noexcept
 }
 
 void
-mpdclient_connected_callback() noexcept
+Instance::OnMpdConnected() noexcept
 {
 #ifndef NCMPC_MINI
 	/* quit if mpd is pre 0.14 - song id not supported by mpd */
-	auto &client = global_instance->GetClient();
 	auto *connection = client.GetConnection();
 	if (mpd_connection_cmp_server_version(connection, 0, 21, 0) < 0) {
 		const unsigned *version =
@@ -129,51 +126,43 @@ mpdclient_connected_callback() noexcept
 		doupdate();
 
 		/* try again after 30 seconds */
-		global_instance->ScheduleReconnect(std::chrono::seconds(30));
+		ScheduleReconnect(std::chrono::seconds{30});
 		return;
 	}
 #endif
 
-	screen->status_bar.ClearMessage();
+	screen_manager.status_bar.ClearMessage();
 	doupdate();
 
-	global_instance->UpdateClient();
+	UpdateClient();
 
-	auto_update_timer(*global_instance);
+	auto_update_timer(*this);
 }
 
 void
-mpdclient_failed_callback() noexcept
+Instance::OnMpdConnectFailed() noexcept
 {
 	/* try again in 5 seconds */
-	global_instance->ScheduleReconnect(std::chrono::seconds(5));
+	ScheduleReconnect(std::chrono::seconds(5));
 }
 
 void
-mpdclient_lost_callback() noexcept
+Instance::OnMpdConnectionLost() noexcept
 {
-	screen->Update(global_instance->GetClient(),
-		       global_instance->GetSeek());
-
-	global_instance->ScheduleReconnect(std::chrono::seconds(1));
+	screen_manager.Update(client, seek);
+	ScheduleReconnect(std::chrono::seconds{1});
 }
 
-/**
- * This function is called by the gidle.c library when MPD sends us an
- * idle event (or when the connection dies).
- */
 void
-mpdclient_idle_callback([[maybe_unused]] unsigned events) noexcept
+Instance::OnMpdIdle([[maybe_unused]] unsigned events) noexcept
 {
-	auto &client = global_instance->GetClient();
-
 #ifndef NCMPC_MINI
 	if (options.enable_xterm_title)
 		update_xterm_title(client);
 #endif
 
-	screen->Update(client, global_instance->GetSeek());
-	auto_update_timer(*global_instance);
+	screen_manager.Update(client, seek);
+	auto_update_timer(*this);
 }
 
 void
@@ -189,51 +178,45 @@ Instance::OnUpdateTimer() noexcept
 }
 
 void
-begin_input_event() noexcept
+Instance::OnUpdateScreen() noexcept
 {
-}
-
-void end_input_event() noexcept
-{
-	auto &client = global_instance->GetClient();
-
-	screen->Update(client, global_instance->GetSeek());
+	screen_manager.Update(client, seek);
 	client.events = (enum mpd_idle)0;
 
-	auto_update_timer(*global_instance);
+	auto_update_timer(*this);
 }
 
-bool
-do_input_event(EventLoop &event_loop, Command cmd) noexcept
+void
+Instance::OnCommand(Command cmd) noexcept
 {
+	update_screen_event.Schedule();
+
 	if (cmd == Command::QUIT) {
 		event_loop.Break();
-		return false;
+		return;
 	}
 
 	try {
-		screen->OnCommand(global_instance->GetClient(),
-				  global_instance->GetSeek(), cmd);
+		screen_manager.OnCommand(GetClient(), GetSeek(), cmd);
 	} catch (...) {
 		screen_status_error(std::current_exception());
-		return true;
+		return;
 	}
 
 	if (cmd == Command::VOLUME_UP || cmd == Command::VOLUME_DOWN)
 		/* make sure we don't update the volume yet */
-		global_instance->DisableUpdateTimer();
-
-	return true;
+		DisableUpdateTimer();
 }
 
 #ifdef HAVE_GETMOUSE
 
 void
-do_mouse_event(Point p, mmask_t bstate) noexcept
+Instance::OnMouse(Point p, mmask_t bstate) noexcept
 {
+	update_screen_event.Schedule();
+
 	try {
-		screen->OnMouse(global_instance->GetClient(),
-				global_instance->GetSeek(), p, bstate);
+		screen_manager.OnMouse(GetClient(), GetSeek(), p, bstate);
 	} catch (...) {
 		screen_status_error(std::current_exception());
 	}
@@ -256,7 +239,7 @@ Instance::OnCheckKeyBindings() noexcept
 		   process */
 		return;
 
-	screen_status_message(buf);
+	screen_manager.Alert(buf);
 
 	doupdate();
 
@@ -301,6 +284,9 @@ try {
 	GetGlobalKeyBindings().Check(nullptr, 0);
 #endif
 
+	/* set options from environment variables */
+	options_env();
+
 	/* parse command line options - 2 pass */
 	options_parse(argc, argv);
 
@@ -308,14 +294,13 @@ try {
 
 	/* create the global Instance */
 	Instance instance;
-	global_instance = &instance;
 	screen = &instance.GetScreenManager();
 
 	AtScopeExit() {
 		/* this must be executed after ~Instance(), so we're
 		   using AtScopeExit() to do the trick */
 #ifndef NCMPC_MINI
-		set_xterm_title("");
+		set_xterm_title({});
 #endif
 		printf("\n");
 	};
