@@ -4,14 +4,14 @@
 #include "OutputsPage.hxx"
 #include "Deleter.hxx"
 #include "PageMeta.hxx"
-#include "ListPage.hxx"
-#include "ListRenderer.hxx"
-#include "screen_status.hxx"
-#include "paint.hxx"
+#include "screen.hxx"
 #include "Command.hxx"
-#include "screen_utils.hxx"
 #include "i18n.h"
-#include "mpdclient.hxx"
+#include "page/ListPage.hxx"
+#include "dialogs/TextInputDialog.hxx"
+#include "ui/ListRenderer.hxx"
+#include "ui/paint.hxx"
+#include "client/mpdclient.hxx"
 #include "util/FNVHash.hxx"
 #include "util/StringAPI.hxx"
 
@@ -22,6 +22,8 @@
 
 #include <assert.h>
 
+using std::string_view_literals::operator""sv;
+
 #if LIBMPDCLIENT_CHECK_VERSION(2,18,0)
 
 [[gnu::pure]]
@@ -105,8 +107,8 @@ class OutputsPage final : public ListPage, ListRenderer {
 #endif
 
 public:
-	OutputsPage(ScreenManager &_screen, WINDOW *w, Size size)
-		:ListPage(w, size), screen(_screen) {}
+	OutputsPage(ScreenManager &_screen, const Window window)
+		:ListPage(_screen, window), screen(_screen) {}
 
 private:
 	void Clear();
@@ -115,7 +117,9 @@ private:
 #if LIBMPDCLIENT_CHECK_VERSION(2,18,0)
 	bool ActivatePartition(struct mpdclient &c,
 			       const struct mpd_partition &partition) noexcept;
-	bool CreateNewPartition(struct mpdclient &c) noexcept;
+
+	[[nodiscard]]
+	Co::InvokeTask CreateNewPartition(struct mpdclient &c) noexcept;
 #endif
 
 	bool Toggle(struct mpdclient &c, unsigned output_index);
@@ -127,10 +131,10 @@ public:
 	void Paint() const noexcept override;
 	void Update(struct mpdclient &c, unsigned events) noexcept override;
 	bool OnCommand(struct mpdclient &c, Command cmd) override;
-	const char *GetTitle(char *s, size_t size) const noexcept override;
+	std::string_view GetTitle(std::span<char> buffer) const noexcept override;
 
 	/* virtual methods from class ListRenderer */
-	void PaintListItem(WINDOW *w, unsigned i, unsigned y, unsigned width,
+	void PaintListItem(Window window, unsigned i, unsigned y, unsigned width,
 			   bool selected) const noexcept override;
 };
 
@@ -151,33 +155,35 @@ OutputsPage::ActivatePartition(struct mpdclient &c,
 		return false;
 	}
 
-	screen_status_printf(_("Switched to partition '%s'"),
-			     partition_name);
+	FmtAlert(_("Switched to partition '{}'"),
+		 partition_name);
 	return true;
 }
 
-inline bool
+inline Co::InvokeTask
 OutputsPage::CreateNewPartition(struct mpdclient &c) noexcept
 {
+	if (!c.IsConnected())
+		co_return;
+
+	const auto name = co_await TextInputDialog{
+		screen, _("Name"),
+	};
+	if (name.empty())
+		co_return;
+
 	auto *connection = c.GetConnection();
 	if (connection == nullptr)
-		return false;
-
-	auto name = screen_readln(screen, _("Name"), nullptr, nullptr, nullptr);
-	if (name.empty())
-		return false;
+		co_return;
 
 	if (!mpd_run_newpartition(connection, name.c_str())) {
 		c.HandleError();
-		return false;
 	}
-
-	return true;
 }
 
 #endif
 
-bool
+inline bool
 OutputsPage::Toggle(struct mpdclient &c, unsigned output_index)
 {
 	if (output_index >= items.size())
@@ -190,7 +196,8 @@ OutputsPage::Toggle(struct mpdclient &c, unsigned output_index)
 		break;
 
 	case Item::Special::NEW_PARTITION:
-		return CreateNewPartition(c);
+		CoStart(CreateNewPartition(c));
+		return true;
 	}
 
 	if (item.partition)
@@ -213,8 +220,8 @@ OutputsPage::Toggle(struct mpdclient &c, unsigned output_index)
 
 		c.events |= MPD_IDLE_OUTPUT;
 
-		screen_status_printf(_("Output '%s' enabled"),
-				     mpd_output_get_name(&output));
+		FmtAlert(_("Output '{}' enabled"),
+			 mpd_output_get_name(&output));
 	} else {
 		if (!mpd_run_disable_output(connection,
 					    mpd_output_get_id(&output))) {
@@ -224,8 +231,8 @@ OutputsPage::Toggle(struct mpdclient &c, unsigned output_index)
 
 		c.events |= MPD_IDLE_OUTPUT;
 
-		screen_status_printf(_("Output '%s' disabled"),
-				     mpd_output_get_name(&output));
+		FmtAlert(_("Output '{}' disabled"),
+			 mpd_output_get_name(&output));
 	}
 
 	return true;
@@ -337,20 +344,20 @@ OutputsPage::Reload(struct mpdclient &c) noexcept
 #endif
 
 	lw.SetLength(items.size());
-	SetDirty();
+	SchedulePaint();
 
 	/* restore the cursor position */
 	lw.SetCursorHash(items, hash);
 }
 
 static std::unique_ptr<Page>
-outputs_init(ScreenManager &screen, WINDOW *w, Size size)
+outputs_init(ScreenManager &screen, const Window window)
 {
-	return std::make_unique<OutputsPage>(screen, w, size);
+	return std::make_unique<OutputsPage>(screen, window);
 }
 
-const char *
-OutputsPage::GetTitle(char *, size_t) const noexcept
+std::string_view 
+OutputsPage::GetTitle(std::span<char>) const noexcept
 {
 	return _("Outputs");
 }
@@ -358,22 +365,22 @@ OutputsPage::GetTitle(char *, size_t) const noexcept
 #if LIBMPDCLIENT_CHECK_VERSION(2,18,0)
 
 static void
-PaintPartition(WINDOW *w, unsigned width, bool selected, bool active,
+PaintPartition(const Window window, unsigned width, bool selected, bool active,
 	       const struct mpd_partition &partition) noexcept
 {
 	const char *name = mpd_partition_get_name(&partition);
 
-	row_color(w, active ? Style::LIST_BOLD : Style::LIST, selected);
-	waddstr(w, _("Partition"));
-	waddstr(w, ": ");
-	waddstr(w, name);
-	row_clear_to_eol(w, width, selected);
+	row_color(window, active ? Style::LIST_BOLD : Style::LIST, selected);
+	window.String(_("Partition"));
+	window.String(": "sv);
+	window.String(name);
+	row_clear_to_eol(window, width, selected);
 }
 
 #endif
 
 void
-OutputsPage::PaintListItem(WINDOW *w, unsigned i,
+OutputsPage::PaintListItem(Window window, unsigned i,
 			   [[maybe_unused]] unsigned y, unsigned width,
 			   bool selected) const noexcept
 {
@@ -386,16 +393,16 @@ OutputsPage::PaintListItem(WINDOW *w, unsigned i,
 		break;
 
 	case Item::Special::NEW_PARTITION:
-		row_color(w, Style::LIST, selected);
-		waddch(w, '[');
-		waddstr(w, _("Create new partition"));
-		waddch(w, ']');
-		row_clear_to_eol(w, width, selected);
+		row_color(window, Style::LIST, selected);
+		window.Char('[');
+		window.String(_("Create new partition"));
+		window.Char(']');
+		row_clear_to_eol(window, width, selected);
 		return;
 	}
 
 	if (item.partition) {
-		PaintPartition(w, width, selected,
+		PaintPartition(window, width, selected,
 			       active_partition == item.GetHash(),
 			       *item.partition);
 		return;
@@ -406,10 +413,10 @@ OutputsPage::PaintListItem(WINDOW *w, unsigned i,
 
 	const auto *output = item.output.get();
 
-	row_color(w, Style::LIST, selected);
-	waddstr(w, mpd_output_get_enabled(output) ? "[X] " : "[ ] ");
-	waddstr(w, mpd_output_get_name(output));
-	row_clear_to_eol(w, width, selected);
+	row_color(window, Style::LIST, selected);
+	window.String(mpd_output_get_enabled(output) ? "[X] "sv : "[ ] "sv);
+	window.String(mpd_output_get_name(output));
+	row_clear_to_eol(window, width, selected);
 }
 
 void
@@ -432,6 +439,8 @@ OutputsPage::Update(struct mpdclient &c, unsigned events) noexcept
 bool
 OutputsPage::OnCommand(struct mpdclient &c, Command cmd)
 {
+	CoCancel();
+
 	if (ListPage::OnCommand(c, cmd))
 		return true;
 
@@ -444,7 +453,7 @@ OutputsPage::OnCommand(struct mpdclient &c, Command cmd)
 		Clear();
 		fill_outputs_list(&c, items);
 		lw.SetLength(items.size());
-		SetDirty();
+		SchedulePaint();
 		return true;
 
 #if LIBMPDCLIENT_CHECK_VERSION(2,18,0)
