@@ -3,23 +3,27 @@
 
 #include "SearchPage.hxx"
 #include "PageMeta.hxx"
+#include "screen.hxx"
 #include "screen_status.hxx"
-#include "TextListRenderer.hxx"
 #include "i18n.h"
 #include "Options.hxx"
 #include "Bindings.hxx"
 #include "GlobalBindings.hxx"
 #include "charset.hxx"
-#include "mpdclient.hxx"
-#include "screen_utils.hxx"
 #include "FileListPage.hxx"
 #include "filelist.hxx"
+#include "dialogs/TextInputDialog.hxx"
+#include "ui/TextListRenderer.hxx"
+#include "lib/fmt/ToSpan.hxx"
+#include "client/mpdclient.hxx"
 #include "util/StringAPI.hxx"
 
 #include <iterator>
 
 #include <string.h>
 
+using std::string_view_literals::operator""sv;
+
 enum {
 	SEARCH_URI = MPD_TAG_COUNT + 100,
 	SEARCH_MODIFIED,
@@ -40,6 +44,7 @@ static constexpr struct {
 	{ MPD_TAG_DATE, "date", N_("date") },
 	{ MPD_TAG_COMPOSER, "composer", N_("composer") },
 	{ MPD_TAG_PERFORMER, "performer", N_("performer") },
+	{ MPD_TAG_LABEL, "label", N_("label") },
 	{ MPD_TAG_COMMENT, "comment", N_("comment") },
 	{ MPD_TAG_COUNT, nullptr, nullptr }
 };
@@ -88,7 +93,8 @@ static const char *const help_text[] = {
 	"		Example: modified:14d (units: s, M, h, d, m, y)",
 	"",
 	"		Available tags: artist, album, title, track,",
-	"		name, genre, date composer, performer, comment, file",
+	"		name, genre, date composer, performer, label,",
+	"		comment, file",
 	"",
 };
 
@@ -99,8 +105,8 @@ class SearchPage final : public FileListPage {
 	std::string pattern;
 
 public:
-	SearchPage(ScreenManager &_screen, WINDOW *_w, Size size) noexcept
-		:FileListPage(_screen, _w, size,
+	SearchPage(ScreenManager &_screen, const Window _window) noexcept
+		:FileListPage(_screen, _screen, _window,
 			      !options.search_format.empty()
 			      ? options.search_format.c_str()
 			      : options.list_format.c_str()) {
@@ -111,40 +117,36 @@ public:
 private:
 	void Clear(bool clear_pattern) noexcept;
 	void Reload(struct mpdclient &c);
-	void Start(struct mpdclient &c);
+
+	[[nodiscard]]
+	Co::InvokeTask Start(struct mpdclient &c);
 
 public:
 	/* virtual methods from class Page */
 	void Paint() const noexcept override;
 	void Update(struct mpdclient &c, unsigned events) noexcept override;
 	bool OnCommand(struct mpdclient &c, Command cmd) override;
-	const char *GetTitle(char *s, size_t size) const noexcept override;
+	std::string_view GetTitle(std::span<char> buffer) const noexcept override;
 };
 
 /* search info */
 class SearchHelpText final : public ListText {
 public:
 	/* virtual methods from class ListText */
-	const char *GetListItemText(char *buffer, size_t size,
-				    unsigned idx) const noexcept override {
+	std::string_view GetListItemText(std::span<char> buffer,
+					 unsigned idx) const noexcept override {
 		assert(idx < std::size(help_text));
 
-		if (idx == 0) {
-			snprintf(buffer, size,
-				 " %s : %s",
-				 GetGlobalKeyBindings().GetKeyNames(Command::SCREEN_SEARCH).c_str(),
-				 "New search");
-			return buffer;
-		}
+		if (idx == 0)
+			return FmtTruncate(buffer, " {} : {}"sv,
+					   GetGlobalKeyBindings().GetKeyNames(Command::SCREEN_SEARCH),
+					   "New search"sv);
 
-		if (idx == 1) {
-			snprintf(buffer, size,
-				 " %s : %s [%s]",
-				 GetGlobalKeyBindings().GetKeyNames(Command::SEARCH_MODE).c_str(),
-				 get_key_description(Command::SEARCH_MODE),
-				 gettext(mode[options.search_mode].label));
-			return buffer;
-		}
+		if (idx == 1)
+			return FmtTruncate(buffer, " {} : {} [{}]"sv,
+					   GetGlobalKeyBindings().GetKeyNames(Command::SEARCH_MODE),
+					   get_key_description(Command::SEARCH_MODE),
+					   my_gettext(mode[options.search_mode].label));
 
 		return help_text[idx];
 	}
@@ -160,14 +162,14 @@ SearchPage::Clear(bool clear_pattern) noexcept
 	if (clear_pattern)
 		pattern.clear();
 
-	SetDirty();
+	SchedulePaint();
 }
 
 static std::unique_ptr<FileList>
 search_simple_query(struct mpd_connection *connection, bool exact_match,
 		    int table, const char *local_pattern)
 {
-	const LocaleToUtf8 filter_utf8(local_pattern);
+	const LocaleToUtf8Z filter_utf8(local_pattern);
 
 	if (table == SEARCH_ARTIST_TITLE) {
 		mpd_command_list_begin(connection, false);
@@ -337,7 +339,7 @@ try {
 	mpd_search_db_songs(connection, false);
 
 	for (size_t i = 0; i < n; i++) {
-		const LocaleToUtf8 value(matchv[i]);
+		const LocaleToUtf8Z value{matchv[i]};
 
 		if (table[i] == SEARCH_URI)
 			mpd_search_add_uri_constraint(connection,
@@ -403,34 +405,35 @@ SearchPage::Reload(struct mpdclient &c)
 
 	screen_browser_sync_highlights(*filelist, c.playlist);
 
-	SetDirty();
+	SchedulePaint();
 }
 
-void
+inline Co::InvokeTask
 SearchPage::Start(struct mpdclient &c)
 {
-	if (!c.IsConnected())
-		return;
+	if (!c.IsReady())
+		co_return;
 
 	Clear(true);
 
-	pattern = screen_readln(screen, _("Search"),
-				nullptr,
-				&search_history,
-				nullptr);
+	pattern = co_await TextInputDialog{
+		screen, _("Search"),
+		{},
+		&search_history,
+	};
 
 	if (pattern.empty()) {
 		lw.Reset();
-		return;
+		co_return;
 	}
 
 	Reload(c);
 }
 
 static std::unique_ptr<Page>
-screen_search_init(ScreenManager &_screen, WINDOW *w, Size size)
+screen_search_init(ScreenManager &_screen, const Window window)
 {
-	return std::make_unique<SearchPage>(_screen, w, size);
+	return std::make_unique<SearchPage>(_screen, window);
 }
 
 void
@@ -443,21 +446,18 @@ SearchPage::Paint() const noexcept
 	}
 }
 
-const char *
-SearchPage::GetTitle(char *str, size_t size) const noexcept
+std::string_view
+SearchPage::GetTitle(std::span<char> buffer) const noexcept
 {
 	if (advanced_search_mode && !pattern.empty())
-		snprintf(str, size, "%s '%s'", _("Search"), pattern.c_str());
+		return FmtTruncate(buffer, "{} '{}'"sv, _("Search"), pattern);
 	else if (!pattern.empty())
-		snprintf(str, size,
-			 "%s '%s' [%s]",
-			 _("Search"),
-			 pattern.c_str(),
-			 gettext(mode[options.search_mode].label));
+		return FmtTruncate(buffer, "{} '{}' [{}]",
+				   _("Search"),
+				   pattern,
+				   my_gettext(mode[options.search_mode].label));
 	else
 		return _("Search");
-
-	return str;
 }
 
 void
@@ -465,7 +465,7 @@ SearchPage::Update(struct mpdclient &c, unsigned events) noexcept
 {
 	if (filelist != nullptr && events & MPD_IDLE_QUEUE) {
 		screen_browser_sync_highlights(*filelist, c.playlist);
-		SetDirty();
+		SchedulePaint();
 	}
 }
 
@@ -477,12 +477,12 @@ SearchPage::OnCommand(struct mpdclient &c, Command cmd)
 		options.search_mode++;
 		if (mode[options.search_mode].label == nullptr)
 			options.search_mode = 0;
-		screen_status_printf(_("Search mode: %s"),
-				     gettext(mode[options.search_mode].label));
+		FmtAlert("{}: {}"sv, _("Search mode"),
+			 my_gettext(mode[options.search_mode].label));
 
 		if (pattern.empty())
 			/* show the new mode in the help text */
-			SetDirty();
+			SchedulePaint();
 		else if (!advanced_search_mode)
 			/* reload only if the new search mode is going
 			   to be considered */
@@ -494,7 +494,7 @@ SearchPage::OnCommand(struct mpdclient &c, Command cmd)
 		return true;
 
 	case Command::SCREEN_SEARCH:
-		Start(c);
+		CoStart(Start(c));
 		return true;
 
 	case Command::CLEAR:
