@@ -3,24 +3,24 @@
 
 #include "SongPage.hxx"
 #include "PageMeta.hxx"
-#include "ListPage.hxx"
-#include "ListText.hxx"
-#include "TextListRenderer.hxx"
-#include "FileBrowserPage.hxx"
-#include "LyricsPage.hxx"
-#include "screen_find.hxx"
 #include "Command.hxx"
 #include "i18n.h"
 #include "screen.hxx"
 #include "charset.hxx"
 #include "time_format.hxx"
-#include "mpdclient.hxx"
+#include "page/ListPage.hxx"
+#include "ui/ListText.hxx"
+#include "ui/TextListRenderer.hxx"
+#include "lib/fmt/ToSpan.hxx"
+#include "client/mpdclient.hxx"
 #include "util/LocaleString.hxx"
 #include "util/StringAPI.hxx"
 #include "util/StringStrip.hxx"
 
 #include <mpd/client.h>
 
+#include <fmt/format.h>
+
 #include <algorithm>
 #include <iterator>
 #include <vector>
@@ -29,6 +29,8 @@
 #include <assert.h>
 #include <time.h>
 
+using std::string_view_literals::operator""sv;
+
 enum {
 	LABEL_LENGTH = MPD_TAG_COUNT,
 	LABEL_PATH,
@@ -63,6 +65,9 @@ static constexpr struct tag_label tag_labels[] = {
 	{ MPD_TAG_TRACK, N_("Track") },
 	{ MPD_TAG_DATE, N_("Date") },
 	{ MPD_TAG_GENRE, N_("Genre") },
+#if LIBMPDCLIENT_CHECK_VERSION(2,17,0)
+	{ MPD_TAG_LABEL, N_("Label") },
+#endif
 	{ MPD_TAG_COMMENT, N_("Comment") },
 	{ LABEL_PATH, N_("Path") },
 	{ LABEL_BITRATE, N_("Bitrate") },
@@ -105,8 +110,8 @@ class SongPage final : public ListPage, ListText {
 	std::vector<std::string> lines;
 
 public:
-	SongPage(ScreenManager &_screen, WINDOW *w, Size size) noexcept
-		:ListPage(w, size),
+	SongPage(ScreenManager &_screen, const Window _window) noexcept
+		:ListPage(_screen, _window),
 		 screen(_screen) {
 		lw.HideCursor();
 	}
@@ -122,31 +127,36 @@ private:
 	 * Appends a line with a fixed width for the label column.
 	 * Handles nullptr strings gracefully.
 	 */
-	void AppendLine(const char *label, const char *value,
+	void AppendLine(const char *label, std::string_view value,
 			unsigned label_col) noexcept;
 
 	void AppendTag(const struct mpd_song *song,
 		       enum mpd_tag_type tag) noexcept;
 	void AddSong(const struct mpd_song *song) noexcept;
 	void AppendStatsLine(enum stats_label label,
-			     const char *value) noexcept;
+			     std::string_view value) noexcept;
 	bool AddStats(struct mpd_connection *connection) noexcept;
 
 public:
 	/* virtual methods from class Page */
 	void OnClose() noexcept override {
 		Clear();
+		ListPage::OnClose();
 	}
 
 	void Paint() const noexcept override;
 	void Update(struct mpdclient &c, unsigned events) noexcept override;
 	bool OnCommand(struct mpdclient &c, Command cmd) override;
-	const char *GetTitle(char *s, size_t size) const noexcept override;
+	std::string_view GetTitle(std::span<char> buffer) const noexcept override;
+
+	const struct mpd_song *GetSelectedSong() const noexcept override {
+		return selected_song != nullptr ? selected_song : played_song;
+	}
 
 private:
 	/* virtual methods from class ListText */
-	const char *GetListItemText(char *buffer, size_t size,
-				    unsigned i) const noexcept override;
+	std::string_view GetListItemText(std::span<char> buffer,
+					 unsigned i) const noexcept override;
 };
 
 void
@@ -164,36 +174,35 @@ SongPage::Clear() noexcept
 	}
 }
 
-const char *
-SongPage::GetListItemText(char *, size_t, unsigned idx) const noexcept
+std::string_view
+SongPage::GetListItemText(std::span<char>, unsigned idx) const noexcept
 {
-	return lines[idx].c_str();
+	return lines[idx];
 }
 
 static std::unique_ptr<Page>
-screen_song_init(ScreenManager &_screen, WINDOW *w, Size size) noexcept
+screen_song_init(ScreenManager &_screen, const Window window) noexcept
 {
 	for (unsigned i = 0; tag_labels[i].label != nullptr; ++i) {
-		unsigned width = StringWidthMB(gettext(tag_labels[i].label));
+		unsigned width = StringWidthMB(my_gettext(tag_labels[i].label));
 		if (width > max_tag_label_width)
 			max_tag_label_width = width;
 	}
 
 	for (unsigned i = 0; i < std::size(stats_labels); ++i) {
 		if (stats_labels[i] != nullptr) {
-			unsigned width = StringWidthMB(gettext(stats_labels[i]));
+			unsigned width = StringWidthMB(my_gettext(stats_labels[i]));
 
 			if (width > max_stats_label_width)
 				max_stats_label_width = width;
 		}
 	}
 
-	return std::make_unique<SongPage>(_screen, w, size);
+	return std::make_unique<SongPage>(_screen, window);
 }
 
-const char *
-SongPage::GetTitle([[maybe_unused]] char *str,
-		   [[maybe_unused]] size_t size) const noexcept
+std::string_view
+SongPage::GetTitle(std::span<char>) const noexcept
 {
 	return _("Song viewer");
 }
@@ -205,11 +214,10 @@ SongPage::Paint() const noexcept
 }
 
 void
-SongPage::AppendLine(const char *label, const char *value_utf8,
+SongPage::AppendLine(const char *label, std::string_view value_utf8,
 		     unsigned label_col) noexcept
 {
 	assert(label != nullptr);
-	assert(value_utf8 != nullptr);
 
 	static constexpr size_t BUFFER_SIZE = 1024;
 	if (label_col >= BUFFER_SIZE - 16)
@@ -220,18 +228,17 @@ SongPage::AppendLine(const char *label, const char *value_utf8,
 	const int value_col = lw.GetWidth() - label_col;
 	/* calculate the number of required linebreaks */
 	const Utf8ToLocale value_locale(value_utf8);
-	const char *value = value_locale.c_str();
+	const std::string_view value = value_locale;
 
-	const char *const value_end = value + strlen(value);
-	const char *value_iter = value;
+	std::string_view value_iter = value;
 
-	while (*value_iter != 0) {
+	while (!value_iter.empty()) {
 		char buffer[BUFFER_SIZE];
 		const char *const buffer_end = buffer + BUFFER_SIZE;
 
 		char *p = buffer;
 		size_t n_space = label_col;
-		if (value_iter == value) {
+		if (value_iter.data() == value.data()) {
 			const size_t label_length = std::min(strlen(label),
 							     BUFFER_SIZE - 16);
 			p = std::copy_n(label, label_length, p);
@@ -246,17 +253,18 @@ SongPage::AppendLine(const char *label, const char *value_utf8,
 		/* skip whitespaces */
 		value_iter = StripLeft(value_iter);
 
-		const char *value_iter_end = AtWidthMB(value_iter,
-						       value_end - value_iter,
-						       value_col);
-		if (value_iter_end == value_iter)
+		auto truncated_value = TruncateAtWidthMB(value_iter, value_col);
+		if (truncated_value.empty())
 			/* not enough room for anything - bail out */
 			break;
 
-		p += snprintf(p, buffer_end - p, "%.*s",
-			      int(value_iter_end - value_iter), value_iter);
-		value_iter = value_iter_end;
+		value_iter = value_iter.substr(truncated_value.size());
 
+		const std::size_t remaining_space = buffer_end - p;
+		if (truncated_value.size() > remaining_space)
+			truncated_value = truncated_value.substr(0, remaining_space);
+
+		p = std::copy(truncated_value.begin(), truncated_value.end(), p);
 		lines.emplace_back(buffer, p);
 	}
 }
@@ -267,7 +275,7 @@ get_tag_label(unsigned tag) noexcept
 {
 	for (unsigned i = 0; tag_labels[i].label != nullptr; ++i)
 		if (tag_labels[i].tag_type == tag)
-			return gettext(tag_labels[i].label);
+			return my_gettext(tag_labels[i].label);
 
 	assert(tag < MPD_TAG_COUNT);
 	return mpd_tag_name((enum mpd_tag_type)tag);
@@ -293,9 +301,8 @@ SongPage::AddSong(const struct mpd_song *song) noexcept
 {
 	assert(song != nullptr);
 
-	char songpos[16];
-	snprintf(songpos, sizeof(songpos), "%d", mpd_song_get_pos(song) + 1);
-	AppendLine(get_tag_label(LABEL_POSITION), songpos,
+	AppendLine(get_tag_label(LABEL_POSITION),
+		   fmt::format_int{mpd_song_get_pos(song) + 1}.c_str(),
 		   max_tag_label_width);
 
 	AppendTag(song, MPD_TAG_ARTIST);
@@ -304,32 +311,24 @@ SongPage::AddSong(const struct mpd_song *song) noexcept
 
 	/* create time string and add it */
 	if (mpd_song_get_duration(song) > 0) {
-		char length[16];
-		format_duration_short(length, sizeof(length),
-				      mpd_song_get_duration(song));
+		char length_buffer[16];
+		const auto length = format_duration_short(length_buffer, mpd_song_get_duration(song));
 
-		const char *value = length;
+		std::string_view value = length;
 
 		char buffer[64];
 
 		if (mpd_song_get_end(song) > 0) {
 			char start[16], end[16];
-			format_duration_short(start, sizeof(start),
-					      mpd_song_get_start(song));
-			format_duration_short(end, sizeof(end),
-					      mpd_song_get_end(song));
 
-			snprintf(buffer, sizeof(buffer), "%s [%s-%s]\n",
-				 length, start, end);
-			value = buffer;
+			value = FmtTruncate(buffer, "{} [{}-{}]"sv, length,
+					    format_duration_short(start, mpd_song_get_start(song)),
+					    format_duration_short(end, mpd_song_get_end(song)));
 		} else if (mpd_song_get_start(song) > 0) {
 			char start[16];
-			format_duration_short(start, sizeof(start),
-					      mpd_song_get_start(song));
 
-			snprintf(buffer, sizeof(buffer), "%s [%s-]\n",
-				 length, start);
-			value = buffer;
+			value = FmtTruncate(buffer, "{} [{}-]"sv, length,
+					    format_duration_short(start, mpd_song_get_start(song)));
 		}
 
 		AppendLine(get_tag_label(LABEL_LENGTH), value,
@@ -342,6 +341,7 @@ SongPage::AddSong(const struct mpd_song *song) noexcept
 	AppendTag(song, MPD_TAG_TRACK);
 	AppendTag(song, MPD_TAG_DATE);
 	AppendTag(song, MPD_TAG_GENRE);
+	AppendTag(song, MPD_TAG_LABEL);
 	AppendTag(song, MPD_TAG_COMMENT);
 
 	AppendLine(get_tag_label(LABEL_PATH), mpd_song_get_uri(song),
@@ -349,9 +349,9 @@ SongPage::AddSong(const struct mpd_song *song) noexcept
 }
 
 void
-SongPage::AppendStatsLine(enum stats_label label, const char *value) noexcept
+SongPage::AppendStatsLine(enum stats_label label, std::string_view value) noexcept
 {
-	AppendLine(gettext(stats_labels[label]), value, max_stats_label_width);
+	AppendLine(my_gettext(stats_labels[label]), value, max_stats_label_width);
 }
 
 bool
@@ -363,27 +363,18 @@ SongPage::AddStats(struct mpd_connection *connection) noexcept
 
 	lines.emplace_back(_("MPD statistics"));
 
-	char buf[64];
-	snprintf(buf, sizeof(buf), "%d",
-		 mpd_stats_get_number_of_artists(mpd_stats));
-	AppendStatsLine(STATS_ARTISTS, buf);
-	snprintf(buf, sizeof(buf), "%d",
-		 mpd_stats_get_number_of_albums(mpd_stats));
-	AppendStatsLine(STATS_ALBUMS, buf);
-	snprintf(buf, sizeof(buf), "%d",
-		 mpd_stats_get_number_of_songs(mpd_stats));
-	AppendStatsLine(STATS_SONGS, buf);
+	AppendStatsLine(STATS_ARTISTS, fmt::format_int{mpd_stats_get_number_of_artists(mpd_stats)}.c_str());
+	AppendStatsLine(STATS_ALBUMS, fmt::format_int{mpd_stats_get_number_of_albums(mpd_stats)}.c_str());
+	AppendStatsLine(STATS_SONGS, fmt::format_int{mpd_stats_get_number_of_songs(mpd_stats)}.c_str());
 
-	format_duration_long(buf, sizeof(buf),
-			     mpd_stats_get_db_play_time(mpd_stats));
+	char buf[64];
+	format_duration_long(buf, mpd_stats_get_db_play_time(mpd_stats));
 	AppendStatsLine(STATS_DBPLAYTIME, buf);
 
-	format_duration_long(buf, sizeof(buf),
-			     mpd_stats_get_play_time(mpd_stats));
+	format_duration_long(buf, mpd_stats_get_play_time(mpd_stats));
 	AppendStatsLine(STATS_PLAYTIME, buf);
 
-	format_duration_long(buf, sizeof(buf),
-			     mpd_stats_get_uptime(mpd_stats));
+	format_duration_long(buf, mpd_stats_get_uptime(mpd_stats));
 	AppendStatsLine(STATS_UPTIME, buf);
 
 	const time_t t = mpd_stats_get_db_update_time(mpd_stats);
@@ -394,37 +385,37 @@ SongPage::AddStats(struct mpd_connection *connection) noexcept
 	return true;
 }
 
-static void
-audio_format_to_string(char *buffer, size_t size,
-		       const struct mpd_audio_format *format) noexcept
+[[nodiscard]] [[gnu::pure]]
+static std::string_view
+audio_format_to_string(std::span<char> buffer,
+		       const struct mpd_audio_format &format) noexcept
 {
-	if (format->bits == MPD_SAMPLE_FORMAT_FLOAT) {
-		snprintf(buffer, size, "%u:f:%u",
-			 format->sample_rate,
-			 format->channels);
-		return;
+	if (format.bits == MPD_SAMPLE_FORMAT_FLOAT) {
+		return FmtTruncate(buffer, "{}:f:{}"sv,
+				   format.sample_rate,
+				   format.channels);
 	}
 
-	if (format->bits == MPD_SAMPLE_FORMAT_DSD) {
-		if (format->sample_rate > 0 &&
-		    format->sample_rate % 44100 == 0) {
+	if (format.bits == MPD_SAMPLE_FORMAT_DSD) {
+		if (format.sample_rate > 0 &&
+		    format.sample_rate % 44100 == 0) {
 			/* use shortcuts such as "dsd64" which implies the
 			   sample rate */
-			snprintf(buffer, size, "dsd%u:%u",
-				 format->sample_rate * 8 / 44100,
-				 format->channels);
-			return;
+
+			return FmtTruncate(buffer, "dsd{}:{}"sv,
+					   format.sample_rate * 8 / 44100,
+					   format.channels);
 		}
 
-		snprintf(buffer, size, "%u:dsd:%u",
-			 format->sample_rate,
-			 format->channels);
-		return;
+		return FmtTruncate(buffer, "{}:dsd:{}"sv,
+				   format.sample_rate,
+				   format.channels);
 	}
 
-	snprintf(buffer, size, "%u:%u:%u",
-		 format->sample_rate, format->bits,
-		 format->channels);
+	return FmtTruncate(buffer, "{}:{}:{}"sv,
+			   format.sample_rate,
+			   format.bits,
+			   format.channels);
 }
 
 void
@@ -458,11 +449,11 @@ SongPage::Update(struct mpdclient &c, unsigned) noexcept
 		lines.emplace_back(_("Currently playing song"));
 		AddSong(played_song);
 
-		if (mpd_status_get_kbit_rate(c.status) > 0) {
+		if (const auto kbit_rate = mpd_status_get_kbit_rate(c.status);
+		    kbit_rate > 0) {
 			char buf[16];
-			snprintf(buf, sizeof(buf), _("%d kbps"),
-				 mpd_status_get_kbit_rate(c.status));
-			AppendLine(get_tag_label(LABEL_BITRATE), buf,
+			AppendLine(get_tag_label(LABEL_BITRATE),
+				   FmtTruncate(buf, _("{} kbps"), kbit_rate),
 				   max_tag_label_width);
 		}
 
@@ -470,8 +461,8 @@ SongPage::Update(struct mpdclient &c, unsigned) noexcept
 			mpd_status_get_audio_format(c.status);
 		if (format) {
 			char buf[32];
-			audio_format_to_string(buf, sizeof(buf), format);
-			AppendLine(get_tag_label(LABEL_FORMAT), buf,
+			AppendLine(get_tag_label(LABEL_FORMAT),
+				   audio_format_to_string(buf, *format),
 				   max_tag_label_width);
 		}
 
@@ -484,7 +475,7 @@ SongPage::Update(struct mpdclient &c, unsigned) noexcept
 		c.HandleError();
 
 	lw.SetLength(lines.size());
-	SetDirty();
+	SchedulePaint();
 }
 
 bool
@@ -493,47 +484,8 @@ SongPage::OnCommand(struct mpdclient &c, Command cmd)
 	if (ListPage::OnCommand(c, cmd))
 		return true;
 
-	switch(cmd) {
-	case Command::LOCATE:
-		if (selected_song != nullptr) {
-			screen_file_goto_song(screen, c, *selected_song);
-			return true;
-		}
-		if (played_song != nullptr) {
-			screen_file_goto_song(screen, c, *played_song);
-			return true;
-		}
-
-		return false;
-
-#ifdef ENABLE_LYRICS_SCREEN
-	case Command::SCREEN_LYRICS:
-		if (selected_song != nullptr) {
-			screen_lyrics_switch(screen, c, *selected_song, false);
-			return true;
-		}
-		if (played_song != nullptr) {
-			screen_lyrics_switch(screen, c, *played_song, true);
-			return true;
-		}
-		return false;
-
-#endif
-
-	case Command::SCREEN_SWAP:
-		if (selected_song != nullptr)
-			screen.Swap(c, selected_song);
-		else
-		// No need to check if this is null - we'd pass null anyway
-			screen.Swap(c, played_song);
-		return true;
-
-	default:
-		break;
-	}
-
-	if (screen_find(screen, lw, cmd, *this)) {
-		SetDirty();
+	if (screen.find_support.Find(lw, *this, cmd)) {
+		SchedulePaint();
 		return true;
 	}
 
