@@ -3,12 +3,12 @@
 
 #include "TagListPage.hxx"
 #include "screen_status.hxx"
-#include "screen_find.hxx"
 #include "FileListPage.hxx"
 #include "Command.hxx"
 #include "i18n.h"
 #include "charset.hxx"
-#include "mpdclient.hxx"
+#include "page/FindSupport.hxx"
+#include "client/mpdclient.hxx"
 #include "util/StringUTF8.hxx"
 
 #include <algorithm>
@@ -16,6 +16,8 @@
 #include <assert.h>
 #include <string.h>
 
+using std::string_view_literals::operator""sv;
+
 TagFilter
 TagListPage::MakeCursorFilter() const noexcept
 {
@@ -40,13 +42,13 @@ CompareUTF8(const std::string &a, const std::string &b)
 	return CollateUTF8(a.c_str(), b.c_str()) < 0;
 }
 
-const char *
-TagListPage::GetListItemText(char *buffer, size_t size,
+std::string_view
+TagListPage::GetListItemText(std::span<char> buffer,
 			     unsigned idx) const noexcept
 {
 	if (parent != nullptr) {
 		if (idx == 0)
-			return "..";
+			return ".."sv;
 
 		--idx;
 	}
@@ -56,7 +58,7 @@ TagListPage::GetListItemText(char *buffer, size_t size,
 
 	assert(idx < values.size());
 
-	return utf8_to_locale(values[idx].c_str(), buffer, size);
+	return utf8_to_locale(values[idx].c_str(), buffer);
 }
 
 static void
@@ -91,6 +93,8 @@ TagListPage::LoadValues(struct mpdclient &c) noexcept
 	/* sort list */
 	std::sort(values.begin(), values.end(), CompareUTF8);
 	lw.SetLength((parent != nullptr) + values.size() + (all_text != nullptr));
+
+	SchedulePaint();
 }
 
 void
@@ -106,13 +110,13 @@ TagListPage::Reload(struct mpdclient &c)
  * to view the tracks of all albums.
  */
 void
-TagListPage::PaintListItem(WINDOW *w, unsigned i,
-			     [[maybe_unused]] unsigned y, unsigned width,
-			     bool selected) const noexcept
+TagListPage::PaintListItem(const Window window, unsigned i,
+			   [[maybe_unused]] unsigned y, unsigned width,
+			   bool selected) const noexcept
 {
 	if (parent != nullptr) {
 		if (i == 0) {
-			screen_browser_paint_directory(w, width, selected,
+			screen_browser_paint_directory(window, width, selected,
 						       "..");
 			return;
 		}
@@ -121,10 +125,10 @@ TagListPage::PaintListItem(WINDOW *w, unsigned i,
 	}
 
 	if (i < values.size())
-		screen_browser_paint_directory(w, width, selected,
-					       Utf8ToLocale(values[i].c_str()).c_str());
+		screen_browser_paint_directory(window, width, selected,
+					       Utf8ToLocale{values[i]});
 	else
-		screen_browser_paint_directory(w, width, selected,
+		screen_browser_paint_directory(window, width, selected,
 					       all_text);
 }
 
@@ -134,10 +138,10 @@ TagListPage::Paint() const noexcept
 	lw.Paint(*this);
 }
 
-const char *
-TagListPage::GetTitle(char *, size_t) const noexcept
+std::string_view
+TagListPage::GetTitle(std::span<char>) const noexcept
 {
-	return title.c_str();
+	return title;
 }
 
 void
@@ -146,7 +150,6 @@ TagListPage::Update(struct mpdclient &c, unsigned events) noexcept
 	if (events & MPD_IDLE_DATABASE) {
 		/* the db has changed -> update the list */
 		Reload(c);
-		SetDirty();
 	}
 }
 
@@ -177,8 +180,8 @@ add_query(struct mpdclient *c, const TagFilter &filter,
 		   the previous level in the filter */
 		text = filter.empty() ? "?" : filter.front().second.c_str();
 
-	screen_status_printf(_("Adding \'%s\' to queue"),
-			     Utf8ToLocale(text).c_str());
+	screen_status_fmt(_("Adding '{}' to queue"),
+			  (std::string_view)Utf8ToLocale{text});
 
 	mpd_search_add_db_songs(connection, true);
 	AddConstraints(connection, filter);
@@ -237,13 +240,11 @@ TagListPage::OnCommand(struct mpdclient &c, Command cmd)
 	case Command::LIST_RFIND:
 	case Command::LIST_FIND_NEXT:
 	case Command::LIST_RFIND_NEXT:
-		screen_find(screen, lw, cmd, *this);
-		SetDirty();
+		CoStart(find_support.Find(lw, *this, cmd));
 		return true;
 
 	case Command::LIST_JUMP:
-		screen_jump(screen, lw, *this, *this);
-		SetDirty();
+		CoStart(find_support.Jump(lw, *this, *this));
 		return true;
 
 	default:
@@ -251,7 +252,7 @@ TagListPage::OnCommand(struct mpdclient &c, Command cmd)
 	}
 
 	if (lw.HandleCommand(cmd)) {
-		SetDirty();
+		SchedulePaint();
 		return true;
 	}
 
@@ -281,7 +282,7 @@ TagListPage::OnMouse(struct mpdclient &c, Point p,
 			HandleSelect(c);
 	}
 
-	SetDirty();
+	SchedulePaint();
 
 	return true;
 }
