@@ -3,17 +3,16 @@
 
 #include "screen_client.hxx"
 #include "screen_status.hxx"
-#include "mpdclient.hxx"
+#include "client/mpdclient.hxx"
 #include "i18n.h"
 #include "charset.hxx"
 
 void
-screen_database_update(struct mpdclient *c, const char *path)
+screen_database_update(struct mpdclient &c, const char *path)
 {
-	assert(c != nullptr);
-	assert(c->IsConnected());
+	assert(c.IsConnected());
 
-	auto *connection = c->GetConnection();
+	auto *connection = c.GetConnection();
 	if (connection == nullptr)
 		return;
 
@@ -24,13 +23,13 @@ screen_database_update(struct mpdclient *c, const char *path)
 		    mpd_connection_clear_error(connection))
 			screen_status_message(_("Database update running"));
 		else
-			c->HandleError();
+			c.HandleError();
 		return;
 	}
 
 	if (path != nullptr && *path != 0) {
 		screen_status_printf(_("Database update of %s started"),
-				     Utf8ToLocale(path).c_str());
+				     Utf8ToLocaleZ{path}.c_str());
 	} else
 		screen_status_message(_("Database update started"));
 }
