@@ -8,10 +8,28 @@
 
 #include <stdarg.h>
 
+void
+screen_status_message(std::string &&msg) noexcept
+{
+	screen->status_bar.SetMessage(std::move(msg));
+}
+
+void
+screen_status_message(std::string_view msg) noexcept
+{
+	screen_status_message(std::string{msg});
+}
+
 void
 screen_status_message(const char *msg) noexcept
 {
-	screen->status_bar.SetMessage(msg);
+	screen_status_message(std::string{msg});
+}
+
+void
+screen_status_vfmt(fmt::string_view format_str, fmt::format_args args) noexcept
+{
+	screen_status_message(fmt::vformat(format_str, std::move(args)));
 }
 
 void
@@ -28,5 +46,5 @@ screen_status_printf(const char *format, ...) noexcept
 void
 screen_status_error(std::exception_ptr e) noexcept
 {
-	screen_status_message(GetFullMessage(std::move(e)).c_str());
+	screen_status_message(GetFullMessage(std::move(e)));
 }
