@@ -4,6 +4,7 @@
 #include "strfsong.hxx"
 #include "charset.hxx"
 #include "time_format.hxx"
+#include "util/CharUtil.hxx"
 #include "util/UriUtil.hxx"
 #include "TagMask.hxx"
 
@@ -13,6 +14,8 @@
 
 #include <string.h>
 
+using std::string_view_literals::operator""sv;
+
 [[gnu::pure]]
 static const char *
 skip(const char *p) noexcept
@@ -40,33 +43,33 @@ skip(const char *p) noexcept
 
 static char *
 CopyString(char *dest, char *const dest_end,
-	   const char *src, size_t length) noexcept
+	   std::string_view src) noexcept
 {
-	if (length >= size_t(dest_end - dest))
-		length = dest_end - dest - 1;
+	if (src.size() >= size_t(dest_end - dest))
+		src = src.substr(0, dest_end - dest - 1);
 
-	return std::copy_n(src, length, dest);
+	return std::copy(src.begin(), src.end(), dest);
 }
 
 static char *
 CopyStringFromUTF8(char *dest, char *const dest_end,
-		   const char *src_utf8) noexcept
+		   std::string_view src_utf8) noexcept
 {
-	return CopyUtf8ToLocale(dest, dest_end - dest, src_utf8);
+	return CopyUtf8ToLocale({dest, dest_end}, src_utf8);
 }
 
 static char *
 CopyTag(char *dest, char *const end,
-	const struct mpd_song *song, enum mpd_tag_type tag) noexcept
+	const struct mpd_song &song, enum mpd_tag_type tag) noexcept
 {
-	const char *value = mpd_song_get_tag(song, tag, 0);
+	const char *value = mpd_song_get_tag(&song, tag, 0);
 	if (value == nullptr)
 		return dest;
 
 	dest = CopyStringFromUTF8(dest, end, value);
 
 	for (unsigned i = 1; dest + 5 < end &&
-		     (value = mpd_song_get_tag(song, tag, i)) != nullptr;
+		     (value = mpd_song_get_tag(&song, tag, i)) != nullptr;
 	     ++i) {
 		*dest++ = ',';
 		*dest++ = ' ';
@@ -79,7 +82,7 @@ CopyTag(char *dest, char *const end,
 static size_t
 _strfsong(char *const s0, char *const end,
 	  const char *format,
-	  const struct mpd_song *song,
+	  const struct mpd_song &song,
 	  const char **last) noexcept
 {
 	bool found = false;
@@ -87,12 +90,6 @@ _strfsong(char *const s0, char *const end,
 	   found==true instead of found==false. */
 	bool missed = false;
 
-
-	if (song == nullptr) {
-		s0[0] = '\0';
-		return 0;
-	}
-
 	char *s = s0;
 	const char *p;
 	for (p = format; *p != '\0' && s < end - 1;) {
@@ -158,65 +155,66 @@ _strfsong(char *const s0, char *const end,
 
 		/* find the extent of this format specifier (stop at \0, ' ', or esc) */
 		const char *name_end = p + 1;
-		while (*name_end >= 'a' && *name_end <= 'z')
+		while (IsLowerAlphaASCII(*name_end))
 			++name_end;
+		const std::string_view name{p + 1, name_end};
 		size_t n = name_end - p + 1;
 
-		const char *value = nullptr, *value_utf8 = nullptr;
+		std::string_view value{}, value_utf8{};
 		enum mpd_tag_type tag = MPD_TAG_UNKNOWN;
 		bool short_tag = false;
 		char buffer[32];
 
 		if (*name_end != '%')
 			n--;
-		else if (strncmp("%file%", p, n) == 0)
-			value_utf8 = mpd_song_get_uri(song);
-		else if (strncmp("%artist%", p, n) == 0)
+		else if (name == "file"sv)
+			value_utf8 = mpd_song_get_uri(&song);
+		else if (name == "artist"sv)
 			tag = MPD_TAG_ARTIST;
-		else if (strncmp("%albumartist%", p, n) == 0)
+		else if (name == "albumartist"sv)
 			tag = MPD_TAG_ALBUM_ARTIST;
-		else if (strncmp("%composer%", p, n) == 0)
+		else if (name == "composer"sv)
 			tag = MPD_TAG_COMPOSER;
-		else if (strncmp("%performer%", p, n) == 0)
+		else if (name == "performer"sv)
 			tag = MPD_TAG_PERFORMER;
 #if LIBMPDCLIENT_CHECK_VERSION(2,17,0)
-		else if (strncmp("%conductor%", p, n) == 0)
+		else if (name == "conductor"sv)
 			tag = MPD_TAG_CONDUCTOR;
-		else if (strncmp("%work%", p, n) == 0)
+		else if (name == "work"sv)
 			tag = MPD_TAG_WORK;
-		else if (strncmp("%grouping%", p, n) == 0)
+		else if (name == "grouping"sv)
 			tag = MPD_TAG_GROUPING;
+		else if (strncmp("%label%", p, n) == 0)
+			tag = MPD_TAG_LABEL;
 #endif
-		else if (strncmp("%title%", p, n) == 0)
+		else if (name == "title"sv)
 			tag = MPD_TAG_TITLE;
-		else if (strncmp("%album%", p, n) == 0)
+		else if (name == "album"sv)
 			tag = MPD_TAG_ALBUM;
-		else if (strncmp("%shortalbum%", p, n) == 0) {
+		else if (name == "shortalbum"sv) {
 			tag = MPD_TAG_ALBUM;
 			short_tag = true;
 		}
-		else if (strncmp("%track%", p, n) == 0)
+		else if (name == "track"sv)
 			tag = MPD_TAG_TRACK;
-		else if (strncmp("%disc%", p, n) == 0)
+		else if (name == "disc"sv)
 			tag = MPD_TAG_DISC;
-		else if (strncmp("%name%", p, n) == 0)
+		else if (name == "name"sv)
 			tag = MPD_TAG_NAME;
-		else if (strncmp("%date%", p, n) == 0)
+		else if (name == "date"sv)
 			tag = MPD_TAG_DATE;
-		else if (strncmp("%genre%", p, n) == 0)
+		else if (name == "genre"sv)
 			tag = MPD_TAG_GENRE;
-		else if (strncmp("%shortfile%", p, n) == 0) {
-			const char *uri = mpd_song_get_uri(song);
+		else if (name == "shortfile"sv) {
+			const char *uri = mpd_song_get_uri(&song);
 			if (strstr(uri, "://") == nullptr)
 				uri = GetUriFilename(uri);
 			value_utf8 = uri;
-		} else if (strncmp("%time%", p, n) == 0) {
-			unsigned duration = mpd_song_get_duration(song);
+		} else if (name == "time"sv) {
+			unsigned duration = mpd_song_get_duration(&song);
 
 			if (duration > 0)  {
-				format_duration_short(buffer, sizeof(buffer),
-						      duration);
-				value = buffer;
+				value = format_duration_short(buffer, duration);
 			}
 		}
 
@@ -237,27 +235,22 @@ _strfsong(char *const s0, char *const end,
 			continue;
 		}
 
-		if (value_utf8 != nullptr) {
+		if (value_utf8.data() != nullptr) {
 			found = true;
 			s = CopyStringFromUTF8(s, end, value_utf8);
 			continue;
 		}
 
-		size_t value_length;
-
-		if (value == nullptr) {
+		if (value.data() == nullptr) {
 			/* just pass-through any unknown specifiers (including esc) */
 			value = p;
-			value_length = n;
 
 			missed = true;
 		} else {
-			value_length = strlen(value);
-
 			found = true;
 		}
 
-		s = CopyString(s, end, value, value_length);
+		s = CopyString(s, end, value);
 	}
 
 	if(last) *last = p;
@@ -266,11 +259,12 @@ _strfsong(char *const s0, char *const end,
 	return s - s0;
 }
 
-size_t
-strfsong(char *s, size_t max, const char *format,
-	 const struct mpd_song *song) noexcept
+std::string_view
+strfsong(std::span<char> buffer, const char *format,
+	 const struct mpd_song &song) noexcept
 {
-	return _strfsong(s, s + max, format, song, nullptr);
+	std::size_t length = _strfsong(buffer.data(), buffer.data() + buffer.size(), format, song, nullptr);
+	return {buffer.data(), length};
 }
 
 TagMask
